"""AxImageCropDb field type functions - before, after / insert, update, delete
"""
import os
import shutil
import backend.misc as ax_misc


async def before_update(db_session, field, before_form, tobe_form, action,
                        current_user):
    """
    Reads file from tmp/<file_guid>/<file_name> and sets it as
    field value. It will be inserted to database
    Returns:
        Object: Returns updated value of current field"""
    del before_form, action, current_user, tobe_form, db_session
    file = field.value

    # value contaiins tmp file dict
    if file and 'guid' in file:
        tmp_folder = os.path.join(ax_misc.tmp_root_dir, file['guid'])
        tmp_path = os.path.join(tmp_folder, file['name'])
        with open(tmp_path, 'rb') as img:
            field.value = img.read()
        shutil.rmtree(tmp_folder)
    else:
        field.needs_sql_update = False

    return field.value


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ Do the same as after_update """
    return await before_update(
        db_session, field, before_form, tobe_form, action, current_user)



"""This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Defines manipulation with form AxForm.
All mutations are used in form constructor - create/update/delete for Tab, Field
Query form_data is used in AxForm.vue as main form query.
"""

import uuid
import graphene
# from loguru import logger
# from sqlalchemy.exc import DatabaseError
from backend.model import AxForm, AxField, AxFieldType, \
    AxRoleFieldPermission, AxColumn, AxRole
import backend.model as ax_model
import backend.dialects as ax_dialects
from backend.schemas.types import Form, Field, PositionInput, \
    RoleFieldPermission
import backend.schemas.action_schema as action_schema
import ujson as json


async def set_form_values(db_session, ax_form, row_guid):
    """ Select row from DB and set AxForm.fields values, state and rowGuid

    Args:
        ax_form (AxForm): empty AxForm, without field values
        row_guid (str): Guid or AxNum of row

    Returns:
        AxForm: Form with field values
    """
    # TODO get list of fields that user have permission
    allowed_fields = []
    for field in ax_form.db_fields:
        allowed_fields.append(field)

    result = await ax_dialects.dialect.select_one(
        db_session=db_session,
        form=ax_form,
        fields_list=allowed_fields,
        row_guid=row_guid)

    if result:
        ax_form.current_state_name = result[0]['axState']
        ax_form.row_guid = result[0]['guid']
        # populate each AxField with data
        for field in ax_form.fields:
            if field in allowed_fields:
                field.value = await ax_dialects.dialect.get_value(
                    type_name=field.field_type.value_type,
                    value=result[0][field.db_name])
            if field.is_tab is False and field.is_virtual:
                field.value = result[0][field.field_type.default_db_name]

    return ax_form


class CreateTab(graphene.Mutation):
    """ Creates AxField wich is tab """
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()

    ok = graphene.Boolean()
    field = graphene.Field(Field)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = f"Error in gql mutation - form.schema -> CreateTab"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            name = args.get('name')

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(form_guid)
            ).first()

            ax_field = AxField()
            ax_field.name = name
            ax_field.form_guid = ax_form.guid
            ax_field.is_tab = True
            ax_field.position = len(
                [i for i in ax_form.fields if i.is_tab is True]) + 1
            db_session.add(ax_field)
            db_session.flush()
            ok = True
            return CreateTab(field=ax_field, ok=ok)


class UpdateTab(graphene.Mutation):
    """ Updates AxField witch is tab """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()    # AxField guid
        name = graphene.String()    # new name

    ok = graphene.Boolean()
    field = graphene.Field(Field)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - form_schema -> UpdateTab.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            name = args.get('name')

            ax_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(guid)
            ).first()
            ax_field.name = name
            db_session.flush()
            ok = True
            return UpdateTab(field=ax_field, ok=ok)


class DeleteTab(graphene.Mutation):
    """ Deletes AxField witch is tab """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - form_schema -> DeleteTab.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_field)
            return DeleteTab(deleted=guid, ok=True)


class CreateField(graphene.Mutation):
    """ Creates AxField

    Arguments:
        form_guid (str): Guid of AxForm
        name (str): Name of new field. Default from en.json
        tag (str): Tag of created field type. Example - AxNum or AxString
        positions (List(PositionInput)): List of positions for all fields
            of form. When field added, we must change position of all fields
            that are lower then inserted.
        position (int): Position of created field
        parent (str): Guid of tab (AxField) wich is parent to current field

    Returns:
        field: Created AxField
        permissions: Default admin permissions that were created. Used to
            update vuex store of workflow constructor.
    """
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()
        tag = graphene.String()
        positions = graphene.List(PositionInput)
        position = graphene.Int()
        parent = graphene.String()

    ok = graphene.Boolean()
    field = graphene.Field(Field)
    permissions = graphene.List(RoleFieldPermission)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - form_schema -> CreateTab'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            name = args.get('name')
            tag = args.get('tag')
            position = args.get('position')
            positions = args.get('positions')
            parent = args.get('parent')

            cur_name = None
            cur_db_name = None
            name_is_checked = False
            cur_num = 1

            ax_field_type = db_session.query(AxFieldType).filter(
                AxFieldType.tag == tag
            ).first()

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(form_guid)
            ).first()

            # If db table already have {db_name} column -> add digit to db_name
            while name_is_checked is False:
                error_flag = False
                if cur_num > 1:
                    cur_name = name + " " + str(cur_num)
                else:
                    cur_name = name
                cur_db_name = ax_field_type.default_db_name + str(cur_num)
                for field in ax_form.fields:
                    if field.name == cur_name or field.db_name == cur_db_name:
                        error_flag = True
                if error_flag is True:
                    cur_num = cur_num + 1
                else:
                    name_is_checked = True
                    break

            ax_field = AxField()
            ax_field.name = cur_name
            ax_field.db_name = cur_db_name
            ax_field.form_guid = ax_form.guid
            ax_field.value_type = ax_field_type.value_type
            ax_field.field_type_tag = ax_field_type.tag
            ax_field.options_json = "{}"
            ax_field.position = position
            ax_field.parent = uuid.UUID(parent) if parent is not None else None

            if ax_field_type.is_always_whole_row:
                ax_field.is_whole_row = True

            db_session.add(ax_field)

            if ax_field_type.is_virtual is False:
                await ax_dialects.dialect.add_column(
                    db_session=db_session,
                    table=ax_form.db_name,
                    db_name=ax_field.db_name,
                    type_name=ax_field_type.value_type)

            db_session.flush()

            # Update positions of all fields that are lower then created field
            for field in ax_form.fields:
                for pos in positions:
                    if field.guid == uuid.UUID(pos.guid):
                        current_parent = None
                        if pos.parent != '#':
                            current_parent = uuid.UUID(pos.parent)
                        field.position = pos.position
                        field.parent = current_parent

            # add permission for default admin role on every state
            admin_role = db_session.query(AxRole).filter(
                AxRole.is_admin.is_(True)
            ).filter(AxRole.form_guid == ax_form.guid).first()

            permissions = []
            for state in ax_form.states:
                perm = AxRoleFieldPermission()
                perm.form_guid = ax_form.guid
                perm.state_guid = state.guid
                perm.role_guid = admin_role.guid
                perm.field_guid = ax_field.guid
                perm.read = True
                perm.edit = True
                db_session.add(perm)
                permissions.append(perm)

            db_session.flush()
            ax_schema.init_schema()  # re-create GQL schema

            ok = True
            return CreateField(field=ax_field, permissions=permissions, ok=ok)


class UpdateField(graphene.Mutation):
    """ Updates AxField

    Arguments:
        guid (str): Guid of AxField that needs update
        name (str): New field name
        db_name (str): New db_name of field
        is_required (bool): Is field required
        is_whole_row (bool): Is field always displayed in whole row
        options_json (JSONString): Json wich is transformed to dict by graphene.
            Contains public field options that are passed to vue
        private_options_json (JSONString): Contains options that are visible
            only in python backend actions. Not visible in Vue.

    Returns:
        field (AxField): Created field

    """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        db_name = graphene.String()
        is_required = graphene.Boolean(required=False, default_value=None)
        is_whole_row = graphene.Boolean(required=False, default_value=None)
        options_json = graphene.JSONString(required=False, default_value=None)
        private_options_json = graphene.JSONString(
            required=False, default_value=None)

    ok = graphene.Boolean()
    field = graphene.Field(Field)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - form_schema -> UpdateField.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            name = args.get('name')
            db_name = args.get('db_name')
            is_required = args.get('is_required')
            is_whole_row = args.get('is_whole_row')
            options_json = args.get('options_json')
            private_options_json = args.get('private_options_json')

            schema_needs_update = False

            ax_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(guid)
            ).first()

            if name:
                ax_field.name = name

            if db_name and db_name != ax_field.db_name:
                db_name_error = False
                for field in ax_field.form.fields:
                    if field.db_name == db_name and (
                            field.guid != uuid.UUID(guid)):
                        db_name_error = True

                if db_name_error:
                    db_name = db_name + '_enother'

                await ax_dialects.dialect.rename_column(
                    db_session=db_session,
                    table=ax_field.form.db_name,
                    old_name=ax_field.db_name,
                    new_name=db_name,
                    type_name=ax_field.field_type.value_type
                )

                ax_field.db_name = db_name
                db_session.flush()
                schema_needs_update = True

            if options_json:
                ax_field.options_json = json.dumps(options_json)

            if private_options_json:
                ax_field.private_options_json = json.dumps(
                    private_options_json)

            if is_required is not None:
                ax_field.is_required = is_required

            if is_whole_row is not None:
                if ax_field.field_type.is_always_whole_row:
                    ax_field.is_whole_row = True
                else:
                    ax_field.is_whole_row = is_whole_row

            # db_session.flush()

            if schema_needs_update:
                ax_schema.init_schema()  # re-create GQL schema

            ok = True
            return CreateTab(field=ax_field, ok=ok)


class DeleteField(graphene.Mutation):
    """ Deletes AxField """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - form_schema -> DeleteField.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')

            ax_field = db_session.query(AxField).filter(
                AxField.guid == uuid.UUID(guid)
            ).first()

            db_session.query(AxRoleFieldPermission).filter(
                AxRoleFieldPermission.field_guid == ax_field.guid).delete()
            db_session.query(AxColumn).filter(
                AxColumn.field_guid == ax_field.guid).delete()

            if ax_field.is_tab is False and ax_field.is_virtual is False:
                await ax_dialects.dialect.drop_column(
                    db_session=db_session,
                    table=ax_field.form.db_name,
                    column=ax_field.db_name
                )

            db_session.delete(ax_field)
            ax_schema.init_schema()  # re-create GQL schema
            ok = True
            return DeleteField(deleted=guid, ok=ok)


class ChangeFieldsPositions(graphene.Mutation):
    """Change position and parent of multiple fields"""
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        positions = graphene.List(PositionInput)  # positions of all fields

    ok = graphene.Boolean()
    fields = graphene.List(Field)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - form_schema -> ChangeFieldsPositions.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            positions = args.get('positions')

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(form_guid)
            ).one()

            for field in ax_form.fields:
                for position in positions:
                    if field.guid == uuid.UUID(position.guid):
                        current_parent = None
                        if position.parent != '#':
                            current_parent = uuid.UUID(position.parent)
                        field.parent = current_parent
                        field.position = position.position

            # db_session.flush()
            query = Field.get_query(info)
            field_list = query.filter(
                AxField.form_guid == uuid.UUID(form_guid)
            ).all()
            return ChangeFieldsPositions(fields=field_list, ok=True)


class FormQuery(graphene.ObjectType):
    """Query all data of AxForm and related classes"""
    # all_fields = graphene.List(Field, form_field=graphene.String())
    form = graphene.Field(
        Form,
        db_name=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    form_data = graphene.Field(
        Form,
        db_name=graphene.Argument(type=graphene.String, required=True),
        row_guid=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )

    async def resolve_form(self, info, db_name=None, update_time=None):
        """ Get AxForm by db_name

        Args:
            db_name (str, optional): AxForm db_name. Defaults to None.
            update_time (str, optional): Used to disable gql caching.

        Returns:
            AxForm: AxForm
        """
        del update_time
        err = 'form_schema -> resolve_form'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = Form.get_query(info=info)
            return query.filter(AxForm.db_name == db_name).first()

    async def resolve_form_data(
            self, info, db_name=None, row_guid=None, update_time=None):
        """ Get AxForm by db_name and row guid

        Args:
            db_name (str, optional): AxForm db_name. Defaults to None.
            row_guid (str, optional): Guid of table row or AxNum value.
            update_time (str, optional): Used to disable gql caching.

        Returns:
            AxForm: Form with field values, avaluble actions, state, row_guid
        """
        err = 'Error in gql query - form_schema -> resolve_form_data'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            db_session = info.context['session']
            del update_time  # used to disable gql caching

            query = Form.get_query(info=info)
            ax_form = query.filter(AxForm.db_name == db_name).first()

            if row_guid is not None:
                ax_form = await set_form_values(
                    db_session=db_session, ax_form=ax_form, row_guid=row_guid)

            ax_form.avalible_actions = await action_schema.get_actions(
                form=ax_form,
                current_state=ax_form.current_state_name
            )
            return ax_form


class FormMutations(graphene.ObjectType):
    """Combine all mutations"""
    create_tab = CreateTab.Field()
    update_tab = UpdateTab.Field()
    delete_tab = DeleteTab.Field()
    create_field = CreateField.Field()
    update_field = UpdateField.Field()
    delete_field = DeleteField.Field()
    change_fields_positions = ChangeFieldsPositions.Field()



""" Describes all static GraphQL types
Create GQL type for each SqlAlchemy model. All SqlAlchemy columns are created by
graphene. But if we need additional params, we need to discribe them.
Example - AxForm have row_guid param wich is not Column. To use this param
in GQL query we must discribe row_guid again in Form type.

"""
import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
from backend.model import AxForm, AxField, AxFieldType, AxGrid, AxColumn, \
    AxUser, AxGroup2Users, AxRole, AxState, AxAction, AxAction2Role, \
    AxState2Role, AxRoleFieldPermission, AxRole2Users, GUID
from backend.misc import convert_column_to_string

convert_sqlalchemy_type.register(GUID)(convert_column_to_string)


class Action(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxAction
        interfaces = (relay.Node, )


class Form(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxForm
        interfaces = (relay.Node, )
    row_guid = graphene.String()
    modal_guid = graphene.String()
    avalible_actions = graphene.List(Action)


class FieldType(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxFieldType
        interfaces = (relay.Node, )


class PositionInput(graphene.InputObjectType):
    """Used to store position data"""
    guid = graphene.NonNull(graphene.String)
    position = graphene.NonNull(graphene.Int)
    parent = graphene.NonNull(graphene.String)


class ConsoleMessage(graphene.ObjectType):
    """ - """
    modal_guid = graphene.String()
    text = graphene.String()


class ActionNotifyMessage(graphene.ObjectType):
    """ - """
    form_guid = graphene.String()
    form_icon = graphene.String()
    form_db_name = graphene.String()
    row_guid = graphene.String()
    modal_guid = graphene.String()
    action_guid = graphene.String()
    action_db_name = graphene.String()
    action_icon = graphene.String()


class Field(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxField
        interfaces = (relay.Node, )
    value = graphene.String()


class Grid(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxGrid
        interfaces = (relay.Node, )


class Column(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxColumn
        interfaces = (relay.Node, )


class User(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxUser
        interfaces = (relay.Node, )


class Role(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxRole
        interfaces = (relay.Node, )


class State(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxState
        interfaces = (relay.Node, )


class RoleFieldPermission(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxRoleFieldPermission
        interfaces = (relay.Node, )


class Group2Users(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxGroup2Users
        interfaces = (relay.Node, )


class Action2Role(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxAction2Role
        interfaces = (relay.Node, )


class State2Role(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxState2Role
        interfaces = (relay.Node, )


class Role2Users(SQLAlchemyObjectType):  # pylint: disable=missing-docstring
    class Meta:  # pylint: disable=missing-docstring
        model = AxRole2Users
        interfaces = (relay.Node, )



"""Pubsub module. Google aiopubsub for move info """

import sys
import aiopubsub
from loguru import logger

this = sys.modules[__name__]
hub = None
publisher = None


def init_pubsub():
    """Initiate pubsub module, create default publisher"""
    try:
        this.hub = aiopubsub.Hub()
        this.publisher = aiopubsub.Publisher(
            this.hub, prefix=aiopubsub.Key())
    except Exception:
        logger.exception('Error initiationg pubsub module')
        raise



"""Ax1tom field type functions - before, after / insert, update, delete"""
import uuid
from backend.model import Ax1tomReference


async def after_update(db_session, field, before_form, tobe_form, action,
                       current_user):
    """Python code runs for field before update

    Args:
        field (AxField): Current field. Is used for value and guid.
        before_form (AxForm): form with values of initial row.
            Used only for info. Modification is pointles.
        tobe_form (AxForm): form that will be used for database query.
            You can modify other fields here.
        action (AxAction): Performed action
        current_user (AxUser): Current user

    Sample SQL query:
        SELECT cars.guid, cars.model, cars.price, ref.child_guid, Countrey.name
        FROM Cars
        LEFT JOIN _ax_mtom_references as ref ON Cars.guid = ref.row_guid
        LEFT JOIN Countrey ON Countrey.guid = ref.child_guid;

    Returns:
        Object: Returns updated value of current field"""
    del before_form, action, current_user

    db_session.query(Ax1tomReference).filter(
        Ax1tomReference.field_guid == field.guid
    ).filter(
        Ax1tomReference.row_guid == uuid.UUID(str(tobe_form.row_guid))
    ).delete()

    if field.value:
        for child_guid in field.value:
            new_tom = Ax1tomReference()
            new_tom.form_guid = tobe_form.guid
            new_tom.field_guid = field.guid
            new_tom.row_guid = uuid.UUID(str(tobe_form.row_guid))
            new_tom.child_guid = uuid.UUID(str(child_guid))
            db_session.add(new_tom)

    return field.value


async def after_insert(db_session, field, before_form, tobe_form, action,
                       current_user):
    """ Do the same as after_update """
    return await after_update(
        db_session, field, before_form, tobe_form, action, current_user)


async def after_delete(db_session, field, before_form, tobe_form, action,
                       current_user):
    """ Delete all Ax1tomReference """
    del before_form, action, current_user
    db_session.query(Ax1tomReference).filter(
        Ax1tomReference.field_guid == field.guid
    ).filter(
        Ax1tomReference.row_guid == uuid.UUID(str(tobe_form.row_guid))
    ).delete()
    return field.value



"""Describes schema for AxForm manipulation in admin home
Create/update/delete of Form, Folder """
import uuid
import os
import shutil
from sqlalchemy import MetaData
import graphene
from loguru import logger
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
import ujson as json
from backend.misc import convert_column_to_string  # TODO check if needed
from backend.model import GUID, AxForm, AxField, AxGrid, AxAction, AxState, \
    AxRole, AxRole2Users, AxRoleFieldPermission, AxState2Role, AxAction2Role, \
    Ax1tomReference
import backend.model as ax_model
# import backend.cache as ax_cache
import backend.dialects as ax_dialects
import backend.misc as ax_misc
from backend.schemas.types import Form, PositionInput

convert_sqlalchemy_type.register(GUID)(convert_column_to_string)


async def tom_sync_form(db_session, old_form_db_name, new_form_db_name):
    """ Relation fields such as Ax1to1, Ax1tom and Ax1tomTable are using
    options_json to store db_name of related form. If form db_name is changed,
    the Json's in every relation field must be updated.
    For each relation field - check if options contain old form name and update.

    Args:
        old_form_db_name (str): Old db_name of form
        new_form_db_name (str): New db_name of form
    """
    relation_fields = db_session.query(AxField).filter(
        AxField.field_type_tag.in_(('Ax1to1', 'Ax1tom', 'Ax1tomTable'))
    ).all()
    for field in relation_fields:
        if 'form' in field.options.keys() and 'grid' in field.options.keys():
            if field.options['form'] == old_form_db_name:
                new_options = field.options
                new_options['form'] = new_form_db_name
                field.options_json = json.dumps(new_options)
                db_session.flush()


async def is_db_name_avalible(db_name) -> bool:
    """Check if table with name '_db_name' is already exists in database"""
    try:
        meta = MetaData()
        meta.reflect(bind=ax_model.engine)
        db_names = list(meta.tables.keys())
        return db_name not in db_names
    except Exception:
        logger.exception('Error checking if db_name exists in database.')
        raise


async def create_db_table(db_session, db_name) -> None:
    """Create database table"""
    try:
        await ax_dialects.dialect.create_data_table(
            db_session=db_session,
            db_name=db_name)
    except Exception:
        logger.exception('Error creating new Form. Cant create db table')
        raise


async def create_ax_form(db_session, name, db_name) -> object:
    """Creates AxForm object"""
    err = "home_schema -> create_ax_form"
    with ax_model.try_catch(db_session, err) as db_session:
        ax_form = AxForm()
        ax_form.name = name
        ax_form.db_name = db_name
        ax_form.tom_label = "{{ax_form_name}} - {{ax_num}}"
        ax_form.icon = "dice-d6"  # TODO: move default icon to Vue
        db_session.add(ax_form)
        return ax_form


async def create_default_tab(db_session, ax_form, name) -> None:
    """ Create default AxForm tab """
    err = "home_schema -> create_default_tab"
    with ax_model.try_catch(db_session, err) as db_session:
        ax_field = AxField()
        ax_field.name = name
        ax_field.form_guid = ax_form.guid
        ax_field.is_tab = True
        ax_field.position = 1
        db_session.add(ax_field)


async def create_default_grid(db_session, ax_form, name):
    """Creates default AxGrid"""
    err = "home_schema -> create_default_grid"
    with ax_model.try_catch(db_session, err) as db_session:
        ax_grid = AxGrid()
        ax_grid.name = name
        ax_grid.db_name = 'Default'
        ax_grid.form_guid = ax_form.guid
        ax_grid.position = 1

        default_options = {
            "enableQuickSearch": False,
            "enableFlatMode": False,
            "enableColumnsResize": True,
            "enableFiltering": True,
            "enableSorting": True,
            "enableOpenForm": True,
            "enableActions": True,
            "rowHeight": 45,
            "pinned": 0
        }

        ax_grid.options_json = json.dumps(default_options)
        ax_grid.is_default_view = True
        db_session.add(ax_grid)


async def create_default_states(db_session,
                                ax_form,
                                default_start,
                                default_state,
                                default_all,
                                default_deleted):
    """Creates default AxStates"""
    err = "home_schema -> create_default_states"
    with ax_model.try_catch(db_session, err) as db_session:
        start = AxState()
        start.name = default_start
        start.form_guid = ax_form.guid
        start.x = 0
        start.y = -200
        start.is_start = True
        db_session.add(start)

        created = AxState()
        created.name = default_state
        created.form_guid = ax_form.guid
        created.x = 0
        created.y = 0
        db_session.add(created)

        deleted = AxState()
        deleted.name = default_deleted
        deleted.form_guid = ax_form.guid
        deleted.x = 0
        deleted.y = 200
        deleted.is_deleted = True
        db_session.add(deleted)

        all_state = AxState()
        all_state.name = default_all
        all_state.form_guid = ax_form.guid
        all_state.x = 300
        all_state.y = 20
        all_state.is_all = True
        db_session.add(all_state)

        return {
            "start": start.guid,
            "created": created.guid,
            "deleted": deleted.guid
        }


async def create_default_actions(
        db_session, ax_form, states, default_create, default_delete,
        default_update, delete_confirm):
    """Creates default AxActions"""
    err = "home_schema -> create_default_actions"
    with ax_model.try_catch(db_session, err) as db_session:
        create = AxAction()
        create.name = default_create
        create.form_guid = ax_form.guid
        create.from_state_guid = states['start']
        create.to_state_guid = states['created']
        db_session.add(create)

        update = AxAction()
        update.name = default_update
        update.form_guid = ax_form.guid
        update.from_state_guid = states['created']
        update.to_state_guid = states['created']
        db_session.add(update)

        delete = AxAction()
        delete.name = default_delete
        delete.form_guid = ax_form.guid
        delete.from_state_guid = states['created']
        delete.to_state_guid = states['deleted']
        delete.confirm_text = delete_confirm
        db_session.add(delete)

        return {
            "create": create.guid,
            "delete": delete.guid
        }


async def create_default_roles(db_session, ax_form, states, actions,
                               default_admin):
    "Creates default AxRole"

    err = "home_schema -> create_default_roles"
    with ax_model.try_catch(db_session, err) as db_session:
        admin_role = AxRole()
        admin_role.name = default_admin
        admin_role.form_guid = ax_form.guid
        admin_role.is_admin = True
        # TODO add admin group user to role
        db_session.add(admin_role)

        db_session.flush()

        start_role = AxState2Role()
        start_role.state_guid = states['start']
        start_role.role_guid = admin_role.guid
        db_session.add(start_role)

        created_role = AxState2Role()
        created_role.state_guid = states['created']
        created_role.role_guid = admin_role.guid
        db_session.add(created_role)

        create_action_role = AxAction2Role()
        create_action_role.action_guid = actions['create']
        create_action_role.role_guid = admin_role.guid
        db_session.add(create_action_role)

        delete_action_role = AxAction2Role()
        delete_action_role.action_guid = actions['delete']
        delete_action_role.role_guid = admin_role.guid
        db_session.add(delete_action_role)


class CreateForm(graphene.Mutation):
    """ Creates AxForm """
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()
        db_name = graphene.String()
        default_tab_name = graphene.String()
        default_grid_name = graphene.String()
        default_start = graphene.String()
        default_all = graphene.String()
        default_create = graphene.String()
        default_state = graphene.String()
        default_delete = graphene.String()
        default_deleted = graphene.String()
        default_admin = graphene.String()
        default_update = graphene.String()
        delete_confirm = graphene.String()

    ok = graphene.Boolean()
    avalible = graphene.Boolean()
    form = graphene.Field(Form)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = "home_schema -> CreateForm"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            name = args.get('name')
            db_name = args.get('db_name')
            default_tab_name = args.get('default_tab_name')
            default_grid_name = args.get('default_grid_name')
            default_start = args.get('default_start')
            default_all = args.get('default_all')
            default_create = args.get('default_create')
            default_state = args.get('default_state')
            default_delete = args.get('default_delete')
            default_deleted = args.get('default_deleted')
            default_admin = args.get('default_admin')
            default_update = args.get('default_update')
            delete_confirm = args.get('delete_confirm')

            name_avalible = await is_db_name_avalible(db_name=db_name)
            if name_avalible is False:
                return CreateForm(form=None, avalible=False, ok=True)

            await create_db_table(
                db_session=db_session,
                db_name=db_name)

            new_form = await create_ax_form(
                db_session=db_session,
                name=name,
                db_name=db_name)

            # Add Admins role to object
            states = await create_default_states(
                db_session=db_session,
                ax_form=new_form,
                default_start=default_start,
                default_state=default_state,
                default_all=default_all,
                default_deleted=default_deleted)

            actions = await create_default_actions(
                db_session=db_session,
                ax_form=new_form,
                states=states,
                default_create=default_create,
                default_delete=default_delete,
                default_update=default_update,
                delete_confirm=delete_confirm)

            await create_default_roles(
                db_session=db_session,
                ax_form=new_form,
                states=states,
                actions=actions,
                default_admin=default_admin)

            await create_default_grid(
                db_session=db_session,
                ax_form=new_form,
                name=default_grid_name)

            await create_default_tab(
                db_session=db_session,
                ax_form=new_form,
                name=default_tab_name)

            db_session.flush()
            ax_schema.init_schema()
            # TODO: check for multiple workers and load balancers.
            # https://community.sanicframework.org/t/removing-routes-necessary-functionality/29

            return CreateForm(form=new_form, avalible=True, ok=True)


class UpdateForm(graphene.Mutation):
    """ Update AxForm """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        db_name = graphene.String()
        icon = graphene.String()
        tom_label = graphene.String()

    ok = graphene.Boolean()
    avalible = graphene.Boolean()
    db_name_changed = graphene.String()
    form = graphene.Field(Form)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = "home_schema -> UpdateForm"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(args.get('guid'))
            ).first()
            old_form_db_name = ax_form.db_name
            db_name_changed = None
            tom_sync_needed = False
            schema_reload_needed = False

            if str(args.get('db_name')) != str(ax_form.db_name):
                db_name_changed = args.get('db_name')
                name_avalible = await is_db_name_avalible(
                    db_name=args.get('db_name'))
                if name_avalible is False:
                    return UpdateForm(
                        form=None,
                        avalible=False,
                        db_name_changed=None,
                        ok=True
                    )
                else:
                    await ax_dialects.dialect.rename_table(
                        db_session=db_session,
                        old=ax_form.db_name,
                        new=args.get('db_name'))
                    tom_sync_needed = True
                    schema_reload_needed = True

            ax_form.name = args.get('name')
            ax_form.db_name = args.get('db_name')
            ax_form.icon = args.get('icon')
            ax_form.tom_label = args.get('tom_label')
            db_session.flush()

            if schema_reload_needed:
                ax_schema.init_schema()

            if tom_sync_needed:
                await tom_sync_form(
                    db_session=db_session,
                    old_form_db_name=old_form_db_name,
                    new_form_db_name=ax_form.db_name)

            return UpdateForm(
                form=ax_form,
                avalible=True,
                db_name_changed=db_name_changed,
                ok=True
            )


class DeleteForm(graphene.Mutation):
    """-"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    forms = graphene.List(Form)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "home_schema -> DeleteForm"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(guid)
            ).first()

            if not ax_form:
                return DeleteForm(forms=None, ok=False)

            # s.execute("SET FOREIGN_KEY_CHECKS=0;")

            # Remove objects connected to form - Ax1tomReference, AxRole2Users,
            # AxState2Role, AxAction2Role, AxRoleFieldPermission, AxRole,
            # AxAction, AxState, AxColumn, AxGrid, AxField
            Ax1tomReference.query.filter(
                Ax1tomReference.form_guid == ax_form.guid).delete()

            for role in ax_form.roles:
                AxRole2Users.query.filter(
                    AxRole2Users.role_guid == role.guid).delete()
                AxState2Role.query.filter(
                    AxState2Role.role_guid == role.guid).delete()
                AxAction2Role.query.filter(
                    AxAction2Role.role_guid == role.guid).delete()
                AxRoleFieldPermission.query.filter(
                    AxRoleFieldPermission.role_guid == role.guid).delete()

                db_session.delete(role)

            for action in ax_form.actions:
                db_session.delete(action)

            for state in ax_form.states:
                db_session.delete(state)

            for grid in ax_form.grids:
                for column in grid.columns:
                    db_session.delete(column)
                db_session.delete(grid)

            for field in ax_form.fields:
                db_session.delete(field)

            await ax_dialects.dialect.drop_table(
                db_session=db_session, db_name=ax_form.db_name)

            db_session.delete(ax_form)
            db_session.flush()

            # s.execute("SET FOREIGN_KEY_CHECKS=1;")

            query = Form.get_query(info)  # SQLAlchemy query
            form_list = query.all()

            # Remove all uploaded files for this form
            uploads_path = ax_misc.path('uploads/form_row_field_file')
            form_folder = os.path.join(uploads_path, str(ax_form.guid))
            if os.path.exists(form_folder) is True:
                shutil.rmtree(form_folder)

            return DeleteForm(forms=form_list, ok=True)


class CreateFolder(graphene.Mutation):
    """ Creates AxForm wich is folder """
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()

    ok = graphene.Boolean()
    form = graphene.Field(Form)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "home_schema -> CreateFolder"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            name = args.get('name')
            ax_form = AxForm()
            ax_form.name = name
            ax_form.is_folder = True
            db_session.add(ax_form)
            return CreateFolder(form=ax_form, ok=True)


class UpdateFolder(graphene.Mutation):
    """ Update AxForm wich is folder """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()

    ok = graphene.Boolean()
    form = graphene.Field(Form)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "home_schema -> UpdateFolder"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            name = args.get('name')

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(guid)
            ).first()
            ax_form.name = name
            db_session.flush()
            return CreateFolder(form=ax_form, ok=True)


class DeleteFolder(graphene.Mutation):
    """ Delete AxForm wich is folder """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    forms = graphene.List(Form)
    ok = graphene.Boolean()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "home_schema -> DeleteFolder"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')

            ax_folder = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(guid)
            ).first()

            sub_objects = db_session.query(AxForm).filter(
                AxForm.parent == ax_folder.guid
            ).all()

            for sub in sub_objects:
                sub.parent = None
                sub.position = 1000

            db_session.delete(ax_folder)
            db_session.flush()

            query = Form.get_query(info)  # SQLAlchemy query
            form_list = query.all()
            return DeleteFolder(forms=form_list, ok=True)


class ChangeFormsPositions(graphene.Mutation):
    """Change position and parent fields of multiple forms"""
    class Arguments:  # pylint: disable=missing-docstring
        positions = graphene.List(PositionInput)

    ok = graphene.Boolean()
    forms = graphene.List(Form)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = "home_schema -> ChangeFormsPositions"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            positions = args.get('positions')
            for position in positions:
                db_form = db_session.query(AxForm).filter(
                    AxForm.guid == uuid.UUID(position.guid)
                ).one()
                current_parent = None
                if position.parent != '#':
                    current_parent = uuid.UUID(position.parent)
                db_form.parent = current_parent
                db_form.position = position.position

            db_session.flush()

            query = Form.get_query(info)  # SQLAlchemy query
            form_list = query.all()
            return ChangeFormsPositions(forms=form_list, ok=True)


class HomeQuery(graphene.ObjectType):
    """AxForm queryes"""
    all_forms = graphene.List(
        Form, update_time=graphene.Argument(
            type=graphene.String, required=False)
    )

    async def resolve_all_forms(self, info, update_time):
        """Get all users"""
        del update_time
        err = 'Error in GQL query - resolve_all_forms.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = Form.get_query(info)  # SQLAlchemy query
            form_list = query.all()
            # await ax_cache.cache.set('form_list', form_list)
            return form_list


class HomeMutations(graphene.ObjectType):
    """Contains all AxForm mutations"""
    create_form = CreateForm.Field()
    update_form = UpdateForm.Field()
    delete_form = DeleteForm.Field()
    create_folder = CreateFolder.Field()
    update_folder = UpdateFolder.Field()
    delete_folder = DeleteFolder.Field()
    change_forms_positions = ChangeFormsPositions.Field()



"""baseline

Revision ID: 9b898740929f
Revises: 
Create Date: 2019-08-06 11:41:56.265057

"""
import sys
from alembic import op
import sqlalchemy as sa
from pathlib import Path

root_path = Path(__file__).resolve().parent.parent.parent.parent
sys.path.insert(0, str(root_path))

import backend.model



# revision identifiers, used by Alembic.
revision = '9b898740929f'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""AxChangelog field type functions - before, after / insert, update, delete"""
import time


async def before_update(db_session, field, before_form, tobe_form, action,
                        current_user):
    """Python code runs for field before update

    Args:
        field (AxField): Current field. Is used for value and guid.
        before_form (AxForm): form with values of initial row.
        Used only for info. Modification is pointles.
        tobe_form (AxForm): form that will be used for database query.
        You can modify other fields here.
        action (AxAction): Performed action
        current_user (AxUser): Current user

    Returns:
        Object: Returns updated value of current field"""

    del current_user, db_session

    field_val = []
    changed_fields = []

    if field.value:
        field_val = field.value
        # json.loads(field.value)

    for f_before in before_form.fields:
        for f_after in tobe_form.fields:
            if f_before.db_name == f_after.db_name:
                if str(f_before.value) != str(f_after.value):
                    changed_field = {
                        "name": f_after.name,
                        "db_name": f_after.db_name,
                        "value": f_after.value
                    }
                    changed_fields.append(changed_field)

    # TODO: add user info
    log_obj = {
        "action": {
            "guid": str(action.guid),
            "name": action.name,
            "messages": action.messages
        },
        "changed_fields": changed_fields,
        "user": None,
        "timestamp": str(time.time())
    }
    field_val.append(log_obj)
    return field_val


async def before_insert(db_session, field, before_form, tobe_form, action,
                        current_user):
    """ before insert """
    del before_form, current_user, field, db_session

    field_val = []
    changed_fields = []

    for f_after in tobe_form.fields:
        if f_after.value:
            changed_field = {
                "name": f_after.name,
                "db_name": f_after.db_name,
                "value": f_after.value
            }
            changed_fields.append(changed_field)

    # TODO: add user info
    log_obj = {
        "action": {
            "guid": str(action.guid),
            "name": action.name,
            "messages": action.messages
        },
        "changed_fields": changed_fields,
        "user": None,
        "timestamp": str(time.time())
    }
    field_val.append(log_obj)
    return field_val



"""Ax GraphQL schema
1. Combines all mutations, queryes and subscribtions from schemas folder.
    Creates static part of schema from them.
2. Creates dynamic part of schema. For each AxGrid the dynamic schema is created
    Schema contains all fields of form. Rows are filtered using AxGris server
    filter. The name of schema is FormName+GridName. If grid is default, the
    additional schema is created, named FormName.
"""
import sys
import graphene
# from loguru import logger
from sqlalchemy import literal
from graphene_sqlalchemy import SQLAlchemyConnectionField
import ujson as json
from backend.schemas.users_schema import UsersQuery
from backend.schemas.users_schema import UsersMutations, UsersSubscription
from backend.schemas.home_schema import HomeQuery, HomeMutations
from backend.schemas.form_schema import FormQuery, FormMutations
from backend.schemas.workflow_schema import WorkflowMutations
from backend.schemas.action_schema import ActionQuery, ActionMutations, \
    ActionSubscription
from backend.schemas.grids_schema import GridsQuery, GridsMutations
from backend.schemas.fields_schema import FieldsQuery, FieldsMutations
from backend.schemas.types import Form, FieldType, Field, Grid, \
    Column, User, Role, State, Action, RoleFieldPermission, Group2Users, \
    Action2Role, State2Role, Role2Users, PositionInput

from backend.model import AxForm
import backend.model as ax_model
import backend.dialects as ax_dialects


this = sys.modules[__name__]
schema = None

#  Dict used to determin what graphene type to use for AxFieldType.value_type
type_dictionary = {
    'text': graphene.String,
    'TEXT': graphene.String,
    'VARCHAR(255)': graphene.String,
    'INT': graphene.Int,
    'DECIMAL(65,2)': graphene.Float,
    'BOOL': graphene.Boolean,
    'GUID': graphene.String,
    'JSON': graphene.String,
    'TIMESTAMP': graphene.Int,
    'BLOB': graphene.String
}

# Static GQL types. Are same as SqlAlchemy model.
gql_types = [
    Form,
    FieldType,
    Field,
    Grid,
    Column,
    User,
    Role,
    State,
    Action,
    RoleFieldPermission,
    Group2Users,
    Action2Role,
    State2Role,
    Role2Users,
    PositionInput,
]


class Query(HomeQuery, FormQuery, UsersQuery, GridsQuery,
            ActionQuery, FieldsQuery, graphene.ObjectType):
    """Combines all schemas queryes"""
    forms = SQLAlchemyConnectionField(Form)
    field_types = SQLAlchemyConnectionField(FieldType)
    fields = SQLAlchemyConnectionField(Field)
    grids = SQLAlchemyConnectionField(Grid)
    columns = SQLAlchemyConnectionField(Column)
    users = SQLAlchemyConnectionField(User)
    roles = SQLAlchemyConnectionField(Role)
    states = SQLAlchemyConnectionField(State)
    actions = SQLAlchemyConnectionField(Action)
    role_field_permissions = SQLAlchemyConnectionField(RoleFieldPermission)
    group_to_users = SQLAlchemyConnectionField(Group2Users)


class Mutations(HomeMutations, FormMutations, UsersMutations, GridsMutations,
                WorkflowMutations, FieldsMutations, ActionMutations,
                graphene.ObjectType):
    """Combines all schemas mutations"""


class Subscription(UsersSubscription, ActionSubscription, graphene.ObjectType):
    """Combines all schemas subscription"""


def make_resolver(db_name, type_class):
    """ Dynamicly create resolver for GrapQL query based on
        db_name - db_name of AxForm + db_name of AxGrid.
        or only db_name of AxForm for default view grid

        Example:
        db_name=MyUberFormGridOne
        MyUberForm is form name
        GridOne is grid name

    """

    async def resolver(
            self, info, update_time=None, quicksearch=None, guids=None):
        del update_time
        err = f"Schema -> Resolver for {db_name}"
        with ax_model.try_catch(info.context['session'], err) as db_session:
            # Find AxForm with name that db_name is started with
            ax_form = None
            found_forms = db_session.query(AxForm).filter(
                literal(db_name).startswith(AxForm.db_name)
            ).all()

            ax_grid = {}
            default_grid = {}

            for form in found_forms:
                # iterate all grids to know, what grid to use
                for grid in form.grids:
                    if form.db_name + grid.db_name == db_name:
                        ax_grid = grid
                        ax_form = form
                    if grid.is_default_view and form.db_name == db_name:
                        default_grid = grid
                        ax_form = form

            grid_to_use = ax_grid or default_grid
            if not grid_to_use:
                return None

            grid_options = json.loads(grid_to_use.options_json)
            server_filter = None
            if 'serverFilter' in grid_options:
                server_filter = grid_options['serverFilter']

            # TODO add permission checks
            allowed_fields = []
            for field in ax_form.db_fields:
                allowed_fields.append(field.db_name)

            # select * from table
            # TODO Add paging
            results = await ax_dialects.dialect.select_all(
                db_session=db_session,
                ax_form=ax_form,
                quicksearch=quicksearch,
                server_filter=server_filter,
                guids=guids)

            result_items = []
            for row in results:
                kwargs = {}
                for key, value in row.items():
                    kwargs[key] = value

                result_items.append(type_class(**kwargs))

            return result_items

    resolver.__name__ = 'resolve_%s' % db_name
    return resolver


def init_schema():
    """Initiate GQL schema. Create dynamic part of schema and combine it with
    static part take from schemas folder"""
    error_msg = "Schema -> init_schema. Error initiating GraphQL shcema."
    with ax_model.scoped_session(error_msg) as db_session:
        # Create typeClass based on each AxForm
        this.schema = None
        type_classes = {}
        all_types = gql_types.copy()  # Add dynamic types to GQL schema

        ax_forms = db_session.query(AxForm).all()
        for form in ax_forms:
            for grid in form.grids:
                # For each grid we create Graphene field with name FormGrid
                capital_form_db_name = form.db_name[0].upper(
                ) + form.db_name[1:]
                class_name = capital_form_db_name + grid.db_name
                class_fields = {}
                class_fields['guid'] = graphene.String()
                class_fields['axNum'] = graphene.Int()
                class_fields['axState'] = graphene.String()
                class_fields['axLabel'] = graphene.String()
                class_fields['axIcon'] = graphene.String()

                # Add fields for each field of AxForm
                for field in form.db_fields:
                    field_type = (
                        type_dictionary[field.field_type.value_type])
                    # TODO maybe add label as description?
                    class_fields[field.db_name] = field_type()

                # Create graphene class and append class dict
                graph_class = type(
                    class_name,
                    (graphene.ObjectType,),
                    class_fields,
                    name=class_name,
                    description=form.name
                )
                type_classes[class_name] = graph_class
                all_types.append(graph_class)

                # if grid is default view we add enother class with name
                # Form without Grid name
                if grid.is_default_view is True:
                    default_class_name = form.db_name[0].upper(
                    ) + form.db_name[1:]

                    default_graph_class = type(
                        default_class_name,
                        (graphene.ObjectType,),
                        class_fields,
                        name=default_class_name,
                        description=form.name
                    )
                    type_classes[default_class_name] = graph_class
                    all_types.append(default_graph_class)

        # Dynamicly crate resolvers for each typeClass
        # Iterate throw created classes and create resolver for each
        dynamic_fields = {}
        for key, type_class in type_classes.items():
            dynamic_fields[key] = graphene.List(
                type_class,
                update_time=graphene.Argument(
                    type=graphene.String, required=False),
                quicksearch=graphene.Argument(
                    type=graphene.String, required=False),
                guids=graphene.Argument(
                    type=graphene.String, required=False)
            )

            dynamic_fields['resolve_%s' % key] = make_resolver(
                key, type_class)

        # Combine static schema query and dynamic query
        DynamicQuery = type('DynamicQuery', (Query,), dynamic_fields)

        this.schema = graphene.Schema(
            query=DynamicQuery,
            mutation=Mutations,
            types=all_types,
            subscription=Subscription
        )



""" This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Defines manipulation with Grids - create/update/delete of AxGrid and AxColumn
Mutation and queryes used only by constructor. AxGrid uses dynamic GQL schema.
Sea schema.py for more info.
"""
import uuid
import graphene
# from loguru import logger
import ujson as json
from backend.model import AxForm, AxGrid, AxColumn, AxField
import backend.model as ax_model


# import backend.cache as ax_cache # TODO use cache!
from backend.schemas.types import Grid, Column, PositionInput


def tom_sync_grid(db_session, form_db_name, old_db_name, new_db_name):
    """ Relation fields such as Ax1to1, Ax1tom and Ax1tomTable are using
    options_json to store db_name of grid used. If grid db_name is changed,
    the Json's in every relation field must be updated.
    For each relation field - check if options contain old form and old grid.

    Args:
        form_db_name (str): Current form db_name
        old_db_name (str): Old db_name of grid
        new_db_name (str): New db_name of grid
    """
    relation_fields = db_session.query(AxField).filter(
        AxField.field_type_tag.in_(('Ax1to1', 'Ax1tom', 'Ax1tomTable'))
    ).all()
    for field in relation_fields:
        if 'form' in field.options.keys() and 'grid' in field.options.keys():
            if field.options['grid'] == old_db_name and \
                    field.options['form'] == form_db_name:

                new_options = field.options
                new_options['grid'] = new_db_name

                field.options_json = json.dumps(new_options)
                db_session.flush()


class CreateColumn(graphene.Mutation):
    """ Creates AxColumn

    Arguments:
        grid_guid (str): Guid of AxGrid
        field_guid (str): Guid of AxField
        column_type (str): Not used!. Was used for responsive grid. #TODO delete
        position (int): Position of column in grid
        positions (List(PositionInput)): Position of other columns in grid

    Returns:
        column (AxColumn): Created grid column
    """
    class Arguments:  # pylint: disable=missing-docstring
        grid_guid = graphene.String()
        field_guid = graphene.String()
        column_type = graphene.String()
        position = graphene.Int()
        positions = graphene.List(PositionInput)

    ok = graphene.Boolean()
    column = graphene.Field(Column)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'grids_schema -> CreateColumn'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            grid_guid = args.get('grid_guid')
            field_guid = args.get('field_guid')
            column_type = args.get('column_type')
            position = args.get('position')
            positions = args.get('positions')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(grid_guid)
            ).first()

            ax_column = AxColumn()
            ax_column.grid_guid = ax_grid.guid
            ax_column.field_guid = uuid.UUID(field_guid)
            ax_column.column_type = column_type
            ax_column.position = position
            db_session.add(ax_column)

            for column in ax_grid.columns:
                for pos in positions:
                    if str(column.guid) == pos.guid:
                        column.position = pos.position
                        column.column_type = pos.parent

            db_session.flush()
            return CreateColumn(column=ax_column, ok=True)


class UpdateColumnOptions(graphene.Mutation):
    """Updates options of column"""
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        options = graphene.String()

    ok = graphene.Boolean()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> UpdateColumnOptions.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            ax_column = db_session.query(AxColumn).filter(
                AxColumn.guid == uuid.UUID(args.get('guid'))
            ).first()
            ax_column.options_json = args.get('options')
            db_session.flush()
            return UpdateColumnOptions(ok=True)


class DeleteColumn(graphene.Mutation):
    """ Deletes AxColumn """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> DeleteColumn'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            ax_column = db_session.query(AxColumn).filter(
                AxColumn.guid == uuid.UUID(guid)
            ).first()
            db_session.delete(ax_column)
            return DeleteColumn(deleted=guid, ok=True)


class ChangeColumnsPositions(graphene.Mutation):
    """Change position and type of columns"""
    class Arguments:  # pylint: disable=missing-docstring
        grid_guid = graphene.String()
        positions = graphene.List(PositionInput)

    ok = graphene.Boolean()
    columns = graphene.List(Column)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - grids_schema -> ChangeColumnsPositions.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            grid_guid = args.get('grid_guid')
            positions = args.get('positions')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(grid_guid)
            ).one()

            for column in ax_grid.columns:
                for pos in positions:
                    if column.guid == uuid.UUID(pos.guid):
                        column.position = pos.position
                        column.column_type = pos.parent
            db_session.flush()
            query = Column.get_query(info)
            column_list = query.filter(
                AxColumn.grid_guid == uuid.UUID(grid_guid)
            ).all()

            ok = True
            return ChangeColumnsPositions(columns=column_list, ok=ok)


class CreateGrid(graphene.Mutation):
    """ Creates AxGrid """
    class Arguments:  # pylint: disable=missing-docstring
        form_guid = graphene.String()
        name = graphene.String()

    ok = graphene.Boolean()
    grid = graphene.Field(Grid)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - grids_schema -> CreateGrid.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            form_guid = args.get('form_guid')
            name = args.get('name')

            ax_form = db_session.query(AxForm).filter(
                AxForm.guid == uuid.UUID(form_guid)
            ).first()

            db_name = "Grid"
            cur_num = 1
            name_is_checked = False
            while name_is_checked is False:
                error_flag = False
                if cur_num > 1:
                    cur_name = name + " " + str(cur_num)
                else:
                    cur_name = name
                cur_db_name = db_name + str(cur_num)
                for grid in ax_form.grids:
                    if grid.name == cur_name or grid.db_name == cur_db_name:
                        error_flag = True
                if error_flag is True:
                    cur_num = cur_num + 1
                else:
                    name_is_checked = True
                    break

            ax_grid = AxGrid()
            ax_grid.name = cur_name
            ax_grid.db_name = db_name + str(cur_num)
            ax_grid.form_guid = ax_form.guid
            ax_grid.position = len(ax_form.grids) + 1

            default_options = {
                "enableQuickSearch": False,
                "enableFlatMode": False,
                "enableColumnsResize": True,
                "enableFiltering": True,
                "enableSorting": True,
                "enableOpenForm": True,
                "enableActions": True,
                "rowHeight": 45,
                "pinned": 0
            }

            ax_grid.options_json = json.dumps(default_options)
            ax_grid.is_default_view = False
            db_session.add(ax_grid)
            db_session.commit()
            ax_schema.init_schema()
            return CreateGrid(grid=ax_grid, ok=True)


class DeleteGrid(graphene.Mutation):
    """ Deletes AxFGrid """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()

    ok = graphene.Boolean()
    deleted = graphene.String()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - grids_schema -> DeleteGrid.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(guid)
            ).first()
            form_db_name = ax_grid.form.name
            old_grid_db_name = ax_grid.db_name
            default_db_name = None

            for grid in ax_grid.form.grids:
                if grid.is_default_view:
                    default_db_name = grid.db_name

            db_session.delete(ax_grid)
            db_session.commit()
            ax_schema.init_schema()
            tom_sync_grid(db_session=db_session,
                          form_db_name=form_db_name,
                          old_db_name=old_grid_db_name,
                          new_db_name=default_db_name)
            return DeleteGrid(deleted=guid, ok=True)


class UpdateGrid(graphene.Mutation):
    """ Update AxGrid """
    class Arguments:  # pylint: disable=missing-docstring
        guid = graphene.String()
        name = graphene.String()
        db_name = graphene.String()
        options_json = graphene.String()
        is_default_view = graphene.Boolean()

    ok = graphene.Boolean()
    grid = graphene.Field(Grid)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        import backend.schema as ax_schema
        err = 'Error in gql mutation - grids_schema -> UpdateGrid.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            guid = args.get('guid')
            name = args.get('name')
            db_name = args.get('db_name')
            options_json = args.get('options_json')
            is_default_view = args.get('is_default_view')
            schema_reload_needed = False
            tom_sync_needed = False

            ax_grid = db_session.query(AxGrid).filter(
                AxGrid.guid == uuid.UUID(guid)
            ).first()
            old_db_name = ax_grid.db_name

            if name:
                ax_grid.name = name

            if db_name:
                ax_grid.db_name = db_name
                schema_reload_needed = True
                if old_db_name != db_name:
                    tom_sync_needed = True

            if options_json:
                ax_grid.options_json = options_json

            if is_default_view:
                all_grids = db_session.query(AxGrid).filter(
                    AxGrid.form_guid == ax_grid.form_guid
                    and AxGrid.guid != ax_grid.guid
                ).all()

                for grid in all_grids:
                    grid.is_default_view = False

                ax_grid.is_default_view = is_default_view
                schema_reload_needed = True

            db_session.commit()

            if schema_reload_needed:
                ax_schema.init_schema()

            if tom_sync_needed:
                tom_sync_grid(
                    db_session=db_session,
                    form_db_name=ax_grid.form.db_name,
                    old_db_name=old_db_name,
                    new_db_name=ax_grid.db_name)

            return UpdateGrid(grid=ax_grid, ok=True)


class GridsQuery(graphene.ObjectType):
    """Query all data of AxGrid and related classes"""
    grid = graphene.Field(
        Grid,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        grid_db_name=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False)
    )
    grids_list = graphene.List(
        Grid,
        form_db_name=graphene.Argument(type=graphene.String, required=True)
    )

    async def resolve_grid(
            self, info, form_db_name, grid_db_name, update_time=None):
        """Get AxGrid by form db_name and grid db_name"""
        del update_time
        err = 'grids_schema -> resolve_grid'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == form_db_name
            ).first()

            query = Grid.get_query(info=info)
            grid = query.filter(AxGrid.form_guid == ax_form.guid).filter(
                AxGrid.db_name == grid_db_name).first()

            # if field is_virtual - we must switch current dbName with
            # field_type.default_db_name
            for column in grid.columns:
                if column.field.field_type.is_virtual:
                    column.field.db_name = (
                        column.field.field_type.default_db_name)

            return grid

    async def resolve_grids_list(self, info, form_db_name):
        """Gets list of all AxGrid's of form """
        err = 'grids_schema -> resolve_grids_list'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == form_db_name
            ).first()

            query = Grid.get_query(info=info)
            grids = query.filter(AxGrid.form_guid == ax_form.guid).all()
            return grids


class GridsMutations(graphene.ObjectType):
    """Combine all mutations"""
    create_column = CreateColumn.Field()
    delete_column = DeleteColumn.Field()
    change_columns_positions = ChangeColumnsPositions.Field()
    create_grid = CreateGrid.Field()
    delete_grid = DeleteGrid.Field()
    update_grid = UpdateGrid.Field()



"""Initiate loguru"""
import sys
from pathlib import Path
from loguru import logger
import backend.misc as ax_misc


def init_logger(logs_filename: str, logs_absolute_path: str, logs_level: str):
    """Initiate loguru

    Args:
        logs_filename (str): Desired name of log file. If None, only console is
            used.
        logs_absolute_path (str): Path where log file must be created. If None -
            the /backend/logs folder will be used
        logs_level (str): What log level to use INFO/ERROR/DEBUG. Default ERROR
    """
    config = {
        "handlers": [
            {
                'sink': sys.stdout,
                'colorize': True,
                'format': '⛏️  | {level} | <level>{message}</level>',
                'backtrace': False
            }
        ],
        "extra": {"user": "someone"}
    }

    try:
        # If log_filename set in app.yaml, then we must write log file
        if logs_filename is not None:
            log_path = ax_misc.path('backend/logs/' + logs_filename)
            if logs_absolute_path is not None:
                log_path = str(Path(logs_absolute_path) / logs_filename)

            # TODO add rotation and retention, compression from app.yaml
            file_handler = {
                'sink': log_path,
                'serialize': True,
                'rotation': '100 MB',
                'enqueue': True,
                'backtrace': False,
                'level': logs_level
            }
            config['handlers'].append(file_handler)

        logger.configure(**config)
    except Exception:
        logger.exception('Error initiating logger.')
        raise



"""Ax workflow apps builder.

Usage:
  ax [--host=<host>] [--port=<port>]
  ax --version
  ax --help

Options:
  -h --help       Show this screen.
  --version       Show version.
  --host=<host>   Host addres to run ax web application (default: 127.0.0.1).
  --port=<port>   Port of ax web application (default: 8080).

"""

import os
import sys
from pathlib import Path
from sanic import Sanic
from sanic_cors import CORS
from loguru import logger
from docopt import docopt

# Add folder with main.py to sys.path.
root_path = Path(__file__).parent.resolve()
if root_path not in sys.path:
    sys.path.insert(0, str(root_path))

# from backend.daemon import daemon
# import daemonocle

# pylint: disable=wrong-import-position
import backend.logger as ax_logger
import backend.misc as ax_misc
import backend.cache as ax_cache
import backend.model as ax_model
import backend.schema as ax_schema
import backend.pubsub as ax_pubsub
import backend.scheduler as ax_scheduler
import backend.migration as ax_migration
import backend.routes as ax_routes
import backend.dialects as ax_dialects

# pylint: enable=wrong-import-position


def init_model():
    """Initiate model. Used in alembic scripts"""
    ax_model.init_model(
        dialect=str(os.environ.get('AX_DB_DIALECT') or 'sqlite'),
        host=str(os.environ.get('AX_DB_HOST') or '127.0.0.1'),
        port=str(os.environ.get('AX_DB_PORT') or '5432'),
        login=str(os.environ.get('AX_DB_LOGIN') or ''),
        password=str(os.environ.get('AX_DB_PASSWORD') or ''),
        database=str(os.environ.get('AX_DB_DATABASE') or 'ax'),
        sqlite_filename=str(os.environ.get(
            'AX_DB_SQLITE_FILENAME') or 'ax_sqlite.db'),
        sqlite_absolute_path=os.environ.get(
            'AX_DB_SQLITE_ABSOLUTE_PATH') or None
    )


# class AxGraphQLView(GraphQLView):
#     """ Extends GraphQLView to output GQL errors"""
#     @staticmethod
#     def format_error(error):
#         logger.error(error)
#         return GraphQLView.format_error(error)


def init_ax():
    """Initiate all modules of Ax"""

    # Load settings from app.yaml to os.environ
    ax_misc.load_configuration()
    # Misc functions used throw all application. Nothing specific.
    ax_misc.init_misc(
        timezone_name=str(os.environ.get('AX_TIMEZONE') or 'UTC'),
        tmp_absolute_path=os.environ.get('AX_TMP_ABSOLUTE_PATH') or None,
        uploads_absolute_path=os.environ.get(
            'AX_UPLOADS_ABSOLUTE_PATH') or None
    )
    # Logger (loguru) used in all modules - console + file + sentry
    ax_logger.init_logger(
        logs_filename=os.environ.get('AX_LOGS_FILENAME') or None,
        logs_absolute_path=os.environ.get('AX_LOGS_ABSOLUTE_PATH') or None,
        logs_level=os.environ.get('AX_LOGS_LEVEL') or 'ERROR'
    )

    # for item, value in os.environ.items():
    #     logger.info('ENV: {}: {}'.format(item, value))

    # Initiate SqlAlchemy. Made with separate function for alembic.
    # It initiates database connection on migration.
    init_model()
    # Init cache module (aiocache). It is not yet used.
    ax_cache.init_cache(
        mode=str(os.environ.get('AX_CACHE_MODE') or 'default'),
        redis_endpoint=str(os.environ.get('AX_REDIS_ENDPOINT') or '127.0.0.1'),
        redis_port=int(os.environ.get('AX_REDIS_PORT') or 6379),
        redis_timeout=int(os.environ.get('AX_REDIS_ENDPOINT') or 1),
    )
    # Initiate pub-sub module. Used for web-socket subscriptions.
    # Notification on workflow action
    ax_pubsub.init_pubsub()
    # Check if database schema needs update
    ax_migration.init_migration()
    # Initiate gql schema.  Depends on cache and pubsub
    ax_schema.init_schema()

    # cors = CORS(app, resources={r"/api/*": {"origins": "*"}})
    CORS(app, automatic_options=True)  # TODO limit CORS to api folder

    @app.listener('before_server_start')
    async def initialize_scheduler(_app, _loop):  # pylint: disable=unused-variable
        """Initiate scheduler. It used for cron-like jobs"""
        ax_scheduler.init_scheduler()

    @app.listener('before_server_start')
    def save_loop(_app, _loop):  # pylint: disable=unused-variable
        """Initiate graphene graphql server"""
        ax_routes.loop = _loop
        ax_routes.app = _app
        ax_routes.init_graphql_view()
        # _app.add_route(ax_routes.graphql_view, '/api/graphql')

    # Initiate all sanic server routes
    ax_routes.init_routes(
        sanic_app=app,
        deck_path=os.environ.get('AX_DECK_ABSOLUTE_PATH')
    )
    # Initiate SQL dialects module. Different SQL queries for differents DBs
    ax_dialects.init_dialects(os.environ.get('AX_DB_DIALECT') or 'sqlite')


ax_logo = """

                                           .
                                          .(
       ..                                 /(#
       ...           --- AX ---          ,/((////.
       ...,,                             ***(/////,
     ****,,,*/                          (****((///*
      ****///////                   ./(((#***/((/*,,,.
      ,/**////////((((/,      ,(((((((((/((**,.,..
        # (///((((((((((((#((((((((((////* ..
          # (((((((((((((##((((((((((((///,
             .#(((((((  ##(((((((((((((///
                       # (###/      /(///
                      # (//,
                      # (#            (//

"""


def main():
    """Main function"""
    arguments = docopt(__doc__)

    host = str(os.environ.get('AX_HOST') or '127.0.0.1')
    port = int(os.environ.get('AX_PORT') or 8080)
    debug = bool(os.environ.get('AX_SANIC_DEBUG') or False)
    access_log = bool(os.environ.get('AX_SANIC_ACCESS_LOG') or False)
    workers = int(os.environ.get('AX_SANIC_WORKERS') or 1)

    if arguments['--host']:
        host = arguments['--host']

    if arguments['--port']:
        port = int(arguments['--port'])

    print(ax_logo)
    logger.info(f'Running Ax on {host}:{port}, {workers} workers running')
    app.run(
        host=host,
        port=port,
        debug=debug,
        access_log=access_log,
        workers=workers)


app = Sanic()
init_ax()

if __name__ == "__main__":
    main()



""" Sanic blueprint for tus protocol - resumable uploads.
Google for TUS protocol fo more info """

import os
import sys
import base64
import uuid
import shutil
from pathlib import Path
from sanic import response
from sanic import Blueprint
from loguru import logger
import ujson as json
import backend.misc as ax_misc

this = sys.modules[__name__]
tus_bp = Blueprint('sanic_tus')
upload_folder = ax_misc.path('tmp')
if ax_misc.server_is_app_engine():
    upload_folder = str(Path('/tmp'))
upload_url = 'upload'
tus_api_version = '1.0.0'
tus_api_version_supported = '1.0.0'
tus_api_extensions = ['creation', 'termination', 'file-check']
tus_max_file_size = 4294967296  # 4GByte


def write_info(data):
    """ write <guid>.info file with upload info"""
    file_guid = data['guid']
    info_path = os.path.join(
        this.upload_folder, file_guid, file_guid + '.info')
    with open(info_path, 'w') as info_file:
        json.dump(data, info_file, indent=4)


def read_info(file_guid):
    """ reads <guid>.info file with upload info"""

    data = None
    info_path = os.path.join(
        this.upload_folder, file_guid, file_guid + '.info')
    if not os.path.exists(info_path):
        raise Exception
    with open(info_path, 'r') as info_file:
        data = json.load(info_file)
    return data


@tus_bp.route('/api/upload', methods=['POST', 'OPTIONS'])
async def tus_file_upload(request):
    """
        OPTION -> return tus information
        POST -> Create empty file with <guid> name and
            <guid>.info wich contain file info
    """

    if request.method == 'OPTIONS' and request.headers.get(
            'Access-Control-Request-Method', None) is not None:
        # CORS option request, return 200, OK
        return response.text("", status=200)

    if request.method == 'OPTIONS':
        return response.text(
            "",
            headers={
                'Tus-Resumable': this.tus_api_version,
                'Tus-Version': this.tus_api_version_supported,
                'Tus-Extension': ",".join(this.tus_api_extensions),
                'Tus-Max-Size': this.tus_max_file_size
            },
            status=204
        )

    if request.headers.get("Tus-Resumable") is None:
        msg = "Received File upload for unsupported file transfer protocol"
        logger.warning(msg)
        return response.text(msg, status=500)  # INTERNAL_SERVER_ERROR

    # We have POST request, do tus CREATION
    metadata = {}
    for kv in request.headers.get("Upload-Metadata", None).split(","):
        (key, value) = kv.split(" ")
        metadata[key] = base64.b64decode(value)

    file_size = int(request.headers.get("Upload-Length", "0"))
    if file_size > this.tus_max_file_size:
        return response.text(msg, status=413)  # REQUEST_ENTITY_TOO_LARGE

    file_guid = str(uuid.uuid4())
    print('-----------------' + str(this.upload_folder))
    content_dir = os.path.join(this.upload_folder, file_guid)
    os.makedirs(content_dir)

    write_info({
        'guid': file_guid,
        'filename': metadata.get("filename"),
        'file_size': file_size,
        'offset': 0,
        'upload-metadata': request.headers.get("Upload-Metadata")
    })

    try:
        content_path = os.path.join(this.upload_folder, file_guid, file_guid)
        with open(content_path, 'w') as content_file:
            content_file.seek(file_size - 1)
            content_file.write("\0")
    except IOError as ex:
        logger.error("Unable to create file: {}".format(ex))
        return response.text("", status=500)  # INTERNAL_SERVER_ERROR

    return response.text(
        "",
        headers={
            'Location': f"{this.upload_url}/{file_guid}"
        },
        status=201
    )  # CREATED


@tus_bp.route('/api/upload/<file_guid>', methods=['HEAD', 'DELETE', 'PATCH'])
def tus_file_upload_chunk(request, file_guid):
    """
        HEAD -> returns current offset
        DELETE -> deletes upload
        PATCH -> add bytes to uploaded file. If filesize is reached -> rename
            uploaded file
    """

    upload_file_path = os.path.join(this.upload_folder, file_guid, file_guid)
    if os.path.lexists(upload_file_path) is False:
        return response.text(
            "",
            headers={
                'Tus-Resumable': this.tus_api_version,
                'Tus-Version': this.tus_api_version_supported
            },
            status=404
        )  # GONE

    file_info = read_info(file_guid=file_guid)

    if request.method == 'HEAD':
        return response.text(
            "",
            headers={
                'Tus-Resumable': this.tus_api_version,
                'Tus-Version': this.tus_api_version_supported,
                'Upload-Offset': file_info['offset'],
                'Cache-Control': 'no-store'
            },
            status=200
        )  # OK

    if request.method == 'DELETE':
        folder = os.path.join(this.upload_folder, file_guid)
        shutil.rmtree(folder)

        return response.text(
            "",
            headers={
                'Tus-Resumable': this.tus_api_version,
            },
            status=204
        )  # NO_CONTENT

    if request.method == 'PATCH':
        filename = file_info['filename']
        file_offset = int(request.headers.get("Upload-Offset", 0))
        chunk_size = int(request.headers.get("Content-Length", 0))
        file_size = int(file_info['file_size'])

        # check to make sure we're in sync
        if file_offset != int(file_info['offset']):
            return response.text("", status=409)  # CONFLICT

        try:
            content_file = open(upload_file_path, "r+b")
        except IOError:
            content_file = open(upload_file_path, "wb")
        finally:
            content_file.seek(file_offset)
            content_file.write(request.body)
            content_file.close()

        new_offset = file_offset + chunk_size
        file_info['offset'] = new_offset
        write_info(file_info)

        # file transfer complete, rename from resource id to actual filename
        if file_size == new_offset:
            os.rename(upload_file_path, os.path.join(
                this.upload_folder, file_guid, filename))
            os.remove(upload_file_path + '.info')

        return response.text(
            "",
            headers={
                'Tus-Resumable': this.tus_api_version,
                'Upload-Offset': new_offset,
            },
            status=204
        )



""" This module is generated by Alembic.
Its runned when Alembic migration in started
"""
from __future__ import with_statement
import sys
from logging.config import fileConfig
from sqlalchemy import engine_from_config
from sqlalchemy import pool
from alembic import context
from loguru import logger
from pathlib import Path

root_path = Path(__file__).resolve().parent.parent.parent
sys.path.insert(0, str(root_path))

import backend.misc as ax_misc
import backend.model as ax_model
import main as ax_app


def include_object(object, name, type_, reflected, compare_to):
    if (type_ == "table" and not str(name).startswith('_')):
        return False
    else:
        return True


print('----------------------------------------------------')

# Sets alembic configuration. migration script distanation and database url
# Database contains alebnic _ax_alembic_version table with current version
alembic_folder = ax_misc.path("backend/alembic")
config = context.config
config.set_main_option('script_location', str(alembic_folder))

fileConfig(config.config_file_name)
ax_misc.load_configuration()
ax_app.init_model()

config.set_main_option('sqlalchemy.url', str(ax_model.db_url))
# target_metadata = ax_model.Base.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    print('running offline Ax migration')
    err = 'env -> run_migrations_offline'
    with ax_model.Session() as db_session:
        url = config.get_main_option("sqlalchemy.url")
        ax_model.Base.query = db_session.query_property()
        target_metadata = ax_model.Base.metadata
        context.configure(
            url=url, target_metadata=target_metadata,
            literal_binds=True,
            include_object=include_object,
            version_table='_ax_alembic_version'
        )
        with context.begin_transaction():
            context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    print('running offline migration')

    connectable = engine_from_config(
        config.get_section(config.config_ini_section),
        prefix="sqlalchemy.",
        poolclass=pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            include_object=include_object,
            target_metadata=target_metadata,
            version_table='_ax_alembic_version'
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



"""Defines Users Scheme and all mutations"""

import asyncio
import graphene
# from graphene import relay
from graphene_sqlalchemy import SQLAlchemyConnectionField
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
import aiopubsub
# from loguru import logger

from backend.misc import convert_column_to_string
from backend.model import AxUser, GUID
import backend.model as ax_model
import backend.cache as ax_cache
import backend.pubsub as ax_pubsub
from backend.schemas.types import User

convert_sqlalchemy_type.register(GUID)(convert_column_to_string)


class CreateUser(graphene.Mutation):
    """ Creates AxUser """
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()
        email = graphene.String()

    ok = graphene.Boolean()
    user = graphene.Field(User)

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - users_schema -> CreateUser.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            new_user = AxUser(
                name=args.get('name'),
                email=args.get('email'),
            )
            db_session.add(new_user)
            db_session.flush()

            ok = True
            ax_pubsub.publisher.publish(aiopubsub.Key('new_user'), new_user)
            return CreateUser(user=new_user, ok=ok)


class MutationExample(graphene.Mutation):
    """Mutation example"""
    class Arguments:  # pylint: disable=missing-docstring
        input_text = graphene.String(required=False, default_value=None)

    output_text = graphene.String()

    def mutate(self, info, input_text=None):  # pylint: disable=missing-docstring
        del info
        ret_text = input_text
        if not input_text:
            ret_text = "default text"
        print('\n\n\n\n!!!' + ret_text + '!!!\n\n\n\n')
        return MutationExample(output_text=ret_text)

# Used to Change Username with Email


class ChangeUserName(graphene.Mutation):
    """Update AxUser"""
    class Arguments:  # pylint: disable=missing-docstring
        name = graphene.String()
        email = graphene.String()

    ok = graphene.Boolean()
    user = graphene.Field(User)

    @classmethod
    def mutate(cls, _, args, context, info):   # pylint: disable=missing-docstring
        err = 'Error in gql - users_schema -> ChangeUserName'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            query = User.get_query(context)
            email = args.get('email')
            name = args.get('name')
            user = query.filter(AxUser.email == email).first()
            user.name = name
            db_session.flush()
            return ChangeUserName(user=user, ok=True)


class UsersQuery(graphene.ObjectType):
    """AxUser queryes"""
    user = SQLAlchemyConnectionField(User)
    find_user = graphene.Field(lambda: User, email=graphene.String())
    all_users = graphene.List(User)
    ping = graphene.String()

    async def resolve_all_users(self, info):
        """Get all users"""
        err = 'Error in GQL query - all_users.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = User.get_query(info)  # SQLAlchemy query
            user_list = query.all()
            await ax_cache.cache.set('user_list', user_list)
            return user_list

    def resolve_find_user(self, args, context, info):
        """default find method"""
        err = 'Error in GQL query - find_user.'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            del info
            query = User.get_query(context)
            email = args.get('email')
            # you can also use and_ with filter()
            # eg: filter(and_(param1, param2)).first()
            return query.filter(AxUser.email == email).first()

    def resolve_ping(self, info):
        """ - """
        return "Pong"


class UsersSubscription(graphene.ObjectType):
    """GraphQL subscriptions"""
    count_seconds = graphene.Float(up_to=graphene.Int())
    mutation_example = graphene.Field(User)
    test_sub = graphene.String()

    async def resolve_count_seconds(self, info, up_to):
        """Test graphql subscription"""
        del info

        for i in range(up_to):
            yield i
            await asyncio.sleep(1.)
        yield up_to

    async def resolve_mutation_example(self, info):
        """Subscribe to adding new user"""
        del info
        try:
            subscriber = aiopubsub.Subscriber(
                ax_pubsub.hub, 'new_user_subscriber')
            subscriber.subscribe(aiopubsub.Key('new_user'))
            while True:
                key, payload = await subscriber.consume()
                del key
                yield payload
        except asyncio.CancelledError:
            await subscriber.remove_all_listeners()


class UsersMutations(graphene.ObjectType):
    """Contains all AxUser mutations"""
    create_user = CreateUser.Field()
    change_username = ChangeUserName.Field()
    mutation_example = MutationExample.Field()



"""AxFiles field type functions - before, after / insert, update, delete"""
import os
import uuid
import shutil
import backend.misc as ax_misc


async def after_update(db_session, field, before_form, tobe_form, action,
                       current_user):
    """
    Moves uploaded files from /tmp/<file_guid>/<file_name> folder to
    /uploads/form_row_field_file/<form_guid>/<row_guid>/<file_guid>/<file_name>
    Returns:
        Object: Returns updated value of current field"""
    del before_form, action, current_user, db_session
    value_guids = []
    form_guid = str(tobe_form.guid)
    row_guid = str(uuid.UUID(str(tobe_form.row_guid)))
    field_guid = str(field.guid)

    if field.value:
        for file in field.value:
            value_guids.append(file['guid'])
            tmp_folder = os.path.join(ax_misc.tmp_root_dir, file['guid'])
            tmp_path = os.path.join(tmp_folder, file['name'])
            dist_folder = os.path.join(
                ax_misc.uploads_root_dir,
                'form_row_field_file',
                form_guid,
                row_guid,
                field_guid,
                file['guid'])
            dist_path = os.path.join(dist_folder, file['name'])

            # if file exists in tmp - move it to row folder
            if os.path.lexists(tmp_path) is True:
                if os.path.exists(dist_folder) is False:
                    os.makedirs(dist_folder)
                shutil.move(tmp_path, dist_path)
                shutil.rmtree(tmp_folder)

    # if form.row.field directory contains sub dirs with guid wich is not
    # in current value -> then file was deleted from field data,
    # We must delete this file from filesystem
    field_folder = os.path.join(
        ax_misc.uploads_root_dir,
        'form_row_field_file',
        form_guid,
        row_guid,
        field_guid
    )
    if os.path.exists(field_folder) is True:
        for root, dirs, _ in os.walk(field_folder):
            del root
            for dir_name in dirs:
                if dir_name not in value_guids:
                    dir_to_delete = os.path.join(field_folder, dir_name)
                    shutil.rmtree(dir_to_delete)

    return field.value


async def after_insert(db_session, field, before_form, tobe_form, action,
                       current_user):
    """ Do the same as after_update """
    return await after_update(
        db_session, field, before_form, tobe_form, action, current_user)


async def after_delete(db_session, field, before_form, tobe_form, action,
                       current_user):
    """
    Deletes all files uploaded for current row
    Returns:
        Object: Returns updated value of current field"""
    del before_form, action, current_user, db_session
    form_guid = str(tobe_form.guid)
    row_guid = str(uuid.UUID(str(tobe_form.row_guid)))
    row_folder = os.path.join(
        ax_misc.uploads_root_dir,
        'form_row_field_file',
        form_guid,
        row_guid
    )

    if os.path.exists(row_folder) is True:
        shutil.rmtree(row_folder)

    return field.value



"""
All SQL queries are generated throw this module.
sea init_dialects
"""
import sys
import sqlite3
import uuid
import re
from loguru import logger
import ujson as json
from backend.model import AxForm
import backend.misc as ax_misc
import backend.model as ax_model

this = sys.modules[__name__]
dialect_name = None
dialect = None


class PorstgreDialect(object):
    """SQL query for Postgre SQL database"""

    name = 'postgre'
    version = None


    async def get_tom_sql(self, form_db_name, form_name, tom_label, fields):
        """    '{{name}} {{surname}}' -> 'Mikhail Marenov'
        Each form have 1toM label option. It could be '{{name}} {{surname}}'
        This function custructs SELECT statement string that will return
        'Mikhail Marenov' string as result. Where 'Mikhail' is from 'name'
        column and 'Marenov' from 'surname' column

        Args:
            form_db_name (str): db_name of form. Same as table name.
            form_name (str): Name of form. Used to replace {{ax_form_name}}
            tom_label (str): Tom label option of AxForm
            fields (List(AxField)): List of all AxFields of current form

        Returns:
            str: SQL statement to be placed in SELECT statement
        """
        tags = re.findall("{{(.*?)}}", tom_label)
        true_tags = []
        for field in fields:
            if field.db_name in tags:
                true_tags.append('"' + field.db_name + '"')

        if "guid" in tags:
            true_tags.append("guid")

        tom_label_modified = tom_label.replace("{{ax_form_name}}", form_name)
        tom_label_modified = tom_label_modified.replace(
            "{{ax_db_name}}", form_db_name)

        texts = re.split('{{.*?}}', tom_label_modified)
        true_texts = []
        for text in texts:
            if text != "":
                true_texts.append("'" + text + "'")
            else:
                true_texts.append(text)

        if not true_tags:
            return f"{true_texts[0]}"

        zip_result = zip(true_texts, true_tags)
        tom_view_arr = [x for item in zip_result for x in item]
        tom_view_arr_clean = [name for name in tom_view_arr if name != ""]

        sql_tom_name = " || ".join(tom_view_arr_clean)
        return sql_tom_name


    async def get_type(self, type_name):
        """Get dialect specific type"""
        postgre_types = {
            'TEXT': 'TEXT',
            'VARCHAR(255)': 'VARCHAR(255)',
            'INT': 'INT',
            'DECIMAL(65,2)': 'DECIMAL(65,2)',
            'BOOL': 'BOOLEAN',
            'GUID': 'UUID',
            'JSON': 'JSON',
            'TIMESTAMP': 'TIMESTAMP',
            'BLOB': 'bytea'
        }
        return postgre_types[type_name]

    async def get_select_sql(self, type_name, db_name):
        """Some value types need a transformation inside SQL Select statement.

        Args:
            type_name (str): value of AxField.AxFieldType.value_type, like INT
            db_name (str): AxField.db_name, the column name of table

        Returns:
            str: SQL statement to be placed in select section
                SET <db_name>=<this function>
        """
        ret_val = None
        if "BLOB" in type_name:
            ret_val = f"octet_length({db_name}) as {db_name}"
        else:
            ret_val = db_name
        return ret_val


    async def get_value_sql(self, type_name, db_name):
        """Some value types need a cast inside SQL.
        Like [UPDATE <table> SET <field>=<this function>]
        It uses param names same as db_name, like :surname

        Args:
            type_name (str): value of AxFieldType.value_type, like INT
            db_name (str): AxField.db_name, the column name of table

        Returns:
            str: SQL statement to be placed in value section
                SET <db_name>=<this function>
        """
        del type_name
        ret_val = None
        # if "TIMESTAMP" in type_name:
        #     ret_val = f"CAST(:{db_name} AS INTEGER)"
        # elif "DECIMAL" in type_name:
        #     ret_val = f"CAST(:{db_name} AS REAL)"
        # else:
        ret_val = f":{db_name}"
        return ret_val


    async def get_value_param(self, type_name, value=None):
        """ Some value types needs to convert field value before submiting
        to SQL.

        Args:
            type_name (str): value of AxFieldType.value_type, like INT
            value (object, optional): Value of field. Defaults to None.

        Returns:
            object: initial value of converted value if needed by field type
        """
        ret_param = None
        if "DECIMAL" in type_name:
            ret_param = str(value).replace(",", ".") if value else None
        elif "JSON" in type_name:
            ret_param = json.dumps(value) if value else None
        else:
            ret_param = value if value else None
        return ret_param


    async def get_value(self, type_name, value):
        """ SELECT cat return different types for diferent dialects
        This method makes returning value same """
        ret_value = value
        if "JSON" in type_name:
            ret_value = json.dumps(value) if value else None
        return ret_value


    def custom_query(self, sql, variables=None):
        """ Executes any SQL. Used in action python code.
        This method is SYNC, leave it so

        Args:
            db_session: SqlAlchemy session
            sql (str): Any sql that needs to be executed
            variables (Dict): Arguments used in sql query

        Returns:
            List(Dict(column_name: value)): result of SQL query
        """

        with ax_model.scoped_session("SQL - custom_query") as db_session:
            res = db_session.execute(sql, variables)
            db_session.commit()
            if res and res.returns_rows:
                return res.fetchall()
            return res


    async def select_all(self, db_session, ax_form, quicksearch=None,
                         server_filter=None, guids=None):
        """ Select * from table

        Args:
            db_session: SqlAlchemy session
            ax_form (AxForm): Current AxForm
            quicksearch (str, optional): Search string from AxGrid.vue
            server_filter (Dict, optional): Dicts with current grid server-
            filter. Constructed by query builder javascript plugin:
                sql (str): Sql expression with params, like 'price > :price'
                params (str): Sql params of query. lile {'price': 20000}
            guids (str, optional): JSON containing list of guids, that must be
                selected. Used in 1tom fields. Where you need to display only
                selected guids.

        Returns:
            List(Dict): Result of SqlAlchemy query. List of rows
        """
        try:
            sql_params = {}

            tom_name = await self.get_tom_sql(
                form_db_name=ax_form.db_name,
                form_name=ax_form.name,
                tom_label=ax_form.tom_label,
                fields=ax_form.db_fields)

            fields_sql = ", ".join(
                '"' + field.db_name + '"' for field in ax_form.db_fields)

            quicksearch_sql = ''
            if quicksearch:
                sql_params['quicksearch'] = quicksearch
                quicksearch_sql = (
                    f"AND {tom_name} LIKE ('%' || :quicksearch || '%') "
                )

            serverfilter_sql = ''
            if server_filter and server_filter["params"]:
                serverfilter_sql = f"AND ({server_filter['sql']})"
                for key, param in server_filter["params"].items():
                    sql_params[key] = param

            guids_sql = ''
            if guids:
                guids_array = json.loads(guids)['items']
                guids_string = ''
                if guids_array:
                    for idx, guid in enumerate(guids_array):
                        del guid
                        guids_array[idx] = guids_array[idx].replace('-', '')

                    for check_guid in guids_array:
                        try:
                            uuid.UUID(check_guid)
                        except Exception:
                            logger.exception(
                                f"Error in guids argument. Cant parse json")
                            raise

                    guids_string = ", ".join(
                        "'" + item + "'" for item in guids_array)
                    guids_sql = f"OR guid IN ({guids_string})"
                    if not quicksearch_sql and not serverfilter_sql:
                        guids_sql = f"AND guid IN ({guids_string})"
            sql = (
                f'SELECT guid, "axState", {fields_sql}'
                f', {tom_name} as "axLabel" FROM "{ax_form.db_name}"'
                f' WHERE (1=1 {quicksearch_sql} {serverfilter_sql}) {guids_sql}'
            )

            result = db_session.execute(sql, sql_params).fetchall()

            return result
            # names = [row[0] for row in result]
        except Exception:
            logger.exception(
                f"Error executing SQL - select_all - {sql}")
            raise



    async def select_one(self, db_session, form, fields_list, row_guid):
        """Select fields from table for one row

        Args:
            db_session: SqlAlchemy session
            form (AxForm): Current form
            fields_list (List(AxField)): Fields that must be selected.
            row_guid (str): String of row guid
            num_fields (List(AxField)): AxNum fields. If not empy, row will
                be searched by guid and every AxNum

        Returns:
            List(Dict(column_name: value)): result of SQL query
        """
        try:
            form_db_name = form.db_name
            num_fields = []
            for field in form.db_fields:
                if field.field_type.tag == 'AxNum':
                    num_fields.append(field)
            fields_string = '"guid", "axState"'
            for field in fields_list:
                field_name = await self.get_select_sql(
                    field.field_type.value_type, field.db_name)
                fields_string += f', "{field_name}"'

            sql = None
            query_params = None

            if ax_misc.string_is_guid(str(row_guid)):
                sql = (f'SELECT {fields_string} '
                       f'FROM "{form_db_name}" '
                       f'WHERE guid=:row_guid')
                query_params = {"row_guid": str(row_guid).replace('-', '')}
            elif num_fields:
                num_where = []
                for num_field in num_fields:
                    num_where.append(f'"{num_field.db_name}"=:ax_num')
                num_where_sql = " OR ".join(num_where)

                sql = (f'SELECT {fields_string} '
                       f'FROM "{form_db_name}" '
                       f'WHERE {num_where_sql}')
                query_params = {"ax_num": str(row_guid)}

            result = db_session.execute(sql, query_params).fetchall()
            return result

        except Exception:
            logger.exception(f"Error executing SQL - select_one - {sql}")
            raise



    async def select_field(self, db_session, form_db_name, field_db_name,
                           row_guid):
        """Gets value of single field from SQL. Used in file viewer.
        See routes.py

        Args:
            db_session: SqlAlchemy session
            form_db_name (str): AxForm.db_name - table name of current form
            fields_db_name (str): Db name of field
            row_guid (str): String of row guid

        Returns:
            List(Dict(column_name: value)): result of SQL query
        """
        try:
            row_guid_stripped = row_guid.replace('-', '')
            sql = (f"SELECT \"{field_db_name}\" "
                   f"FROM {form_db_name} "
                   f"WHERE guid='{row_guid_stripped}'"
                   )
            result = db_session.execute(sql).fetchall()
            return result[0][field_db_name] if result else None
        except Exception:
            logger.exception(
                f"Error executing SQL - select_field {form_db_name}")
            raise


    async def insert(self, db_session, form, to_state_name, new_guid):
        """ Insert row into database and set state with AxForm field values

        Args:
            db_session: SqlAlchemy session
            form (AxForm): Current form with filled field values.
            to_state_name (str): Name of form state that must be set
            new_guid (str): Guid that must be used to create record.

        Returns:
            str: Guid of created row
        """
        try:
            fields_db_names = []
            query_params = {
                "ax_state": to_state_name,
                "row_guid": str(new_guid).replace('-', ''),
            }

            value_strings = []
            for field in form.fields:
                if field.needs_sql_update:
                    fields_db_names.append(f'"{field.db_name}"')
                    value_str = await self.get_value_sql(
                        type_name=field.field_type.value_type,
                        db_name=field.db_name)
                    value_strings.append(value_str)
                    query_params[field.db_name] = await self.get_value_param(
                        type_name=field.field_type.value_type,
                        value=field.value)

            column_sql = ", ".join(fields_db_names)
            values_sql = ", ".join(value_strings)

            sql = (
                f'INSERT INTO "{form.db_name}" '
                f'("guid", "axState", {column_sql}) '
                f"VALUES (:row_guid, :ax_state, {values_sql});"
            )
            db_session.execute(sql, query_params)
            return new_guid
        except Exception:
            logger.exception('Error executing SQL - insert')
            raise


    async def update(self, db_session, form, to_state_name, row_guid):
        """Update database row based of AxForm fields values

        Args:
            db_session: SqlAlchemy session
            form (AxForm): Current form with filled field values.
            to_state_name (str): Name of form state that must be set
            row_guid (str): Guid of updated row

        Returns:
            SqlAlchemy result: Contains nothing. Not used.
        """
        try:
            value_strings = []
            query_params = {
                "ax_state": to_state_name,
                "row_guid": row_guid
            }
            for field in form.fields:
                if field.needs_sql_update:
                    current_valu_str = await self.get_value_sql(
                        type_name=field.field_type.value_type,
                        db_name=field.db_name
                    )
                    value_strings.append(
                        f"\"{field.db_name}\"={current_valu_str}")
                    query_params[field.db_name] = await self.get_value_param(
                        type_name=field.field_type.value_type,
                        value=field.value
                    )

            values_sql = ", ".join(value_strings)
            sql = (
                f"UPDATE \"{form.db_name}\" "
                f"SET \"axState\"=:ax_state, {values_sql} "
                f"WHERE guid=:row_guid "
            )
            return db_session.execute(sql, query_params)

        except Exception:
            logger.exception('Error executing SQL - update')
            raise


    async def delete(self, db_session, form, row_guid):
        """ Delete row of table of form"""
        try:
            sql = (
                f"DELETE FROM \"{form.db_name}\" WHERE guid='{row_guid}' "
            )
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - delete')
            raise


    async def create_data_table(self, db_session, db_name):
        """Create table with system columns"""
        try:
            sql = f"""CREATE TABLE \"{db_name}\" (
                guid UUID PRIMARY KEY,
                \"axState\" VARCHAR NOT NULL
            );"""
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - create_data_table')
            raise

    async def rename_table(self, db_session, old, new):
        """Rename table"""
        try:
            sql = f'ALTER TABLE "{old}" RENAME TO "{new}";'
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - rename_table')
            raise


    async def drop_table(self, db_session, db_name):
        """Drop table"""
        try:
            sql = f'DROP TABLE "{db_name}";'
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - drop_table')
            raise


    async def add_column(self, db_session, table, db_name, type_name):
        """Add column"""
        try:
            dialect_type = await self.get_type(type_name=type_name)
            sql = f'ALTER TABLE "{table}" ADD COLUMN "{db_name}" {dialect_type}'
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - add_column')
            raise

    async def rename_column(self, db_session, table, old_name, new_name,
                            type_name) -> None:
        """Rename column"""
        try:
            del type_name
            sql = (
                f'ALTER TABLE "{table}"'
                f' RENAME COLUMN "{old_name}" TO "{new_name}";'
            )
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - rename_column')
            raise

    async def drop_column(self, db_session, table, column) -> None:
        """ Drop column """
        try:
            sql = f'ALTER TABLE "{table}" DROP COLUMN "{column}";'
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - drop_column')
            raise


class SqliteDialect(PorstgreDialect):
    """SQL query for Sqlite database"""
    name = 'sqlite'
    version = None

    async def get_type(self, type_name) -> str:
        """Get dialect specific type"""
        sqlite_types = {
            'TEXT': 'TEXT',
            'VARCHAR(255)': 'TEXT',
            'INT': 'INTEGER',
            'DECIMAL(65,2)': 'REAL',
            'BOOL': 'NUMERIC',
            'GUID': 'TEXT',
            'TIMESTAMP': 'INTEGER',
            'BLOB': 'BLOB',
            'JSON': 'TEXT'
        }
        return sqlite_types[type_name]

    async def get_select_sql(self, type_name, db_name):
        """Some value types need a transformation inside SQL Select statement.

        Args:
            type_name (str): value of AxField.AxFieldType.value_type, like INT
            db_name (str): AxField.db_name, the column name of table

        Returns:
            str: SQL statement to be placed in select section
                SET <db_name>=<this function>
        """
        ret_val = None
        if "BLOB" in type_name:
            ret_val = f"LENGTH({db_name}) as \"{db_name}\""
        else:
            ret_val = f'"db_name"'
        return ret_val

    async def get_value_sql(self, type_name, db_name):
        """Some value types need a cast inside SQL.
        Like [UPDATE <table> SET <field>=<this function>]
        It uses param names same as db_name, like :surname

        Args:
            type_name (str): value of AxFieldType.value_type, like INT
            db_name (str): AxField.db_name, the column name of table

        Returns:
            str: SQL statement to be placed in value section
                SET <db_name>=<this function>
        """
        ret_val = None
        if "TIMESTAMP" in type_name:
            ret_val = f"CAST(:{db_name} AS INTEGER)"
        elif "DECIMAL" in type_name:
            ret_val = f"CAST(:{db_name} AS REAL)"
        else:
            ret_val = f":{db_name}"
        return ret_val

    async def get_value_param(self, type_name, value=None):
        """ Some value types needs to convert field value before submiting
        to SQL.

        Args:
            type_name (str): value of AxFieldType.value_type, like INT
            value (object, optional): Value of field. Defaults to None.

        Returns:
            object: initial value of converted value if needed by field type
        """
        ret_param = None
        if "DECIMAL" in type_name:
            ret_param = str(value).replace(",", ".") if value else None
        elif "JSON" in type_name:
            ret_param = json.dumps(value) if value else None
        else:
            ret_param = value if value else None
        return ret_param


    async def create_data_table(self, db_session, db_name: str) -> None:
        """Create table with system columns"""
        try:
            sql = f"""CREATE TABLE {db_name} (
                guid VARCHAR PRIMARY KEY,
                axState VARCHAR NOT NULL
            );"""
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - create_data_table')
            raise

    async def rename_column(
            self, db_session, table, old_name, new_name, type_name) -> None:
        """Sqlite allows rename of column only from version 3.25.0.
        We have to reacreate table and copy data"""
        try:
            del type_name
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == table
            ).first()

            drop_tmp = f"DROP TABLE IF EXISTS _ax_tmp"

            create_tmp = (
                f"CREATE TABLE _ax_tmp ("
                f"guid VARCHAR, "
                f"axState VARCHAR"
            )
            for field in ax_form.db_fields:
                db_name = field.db_name if (
                    field.db_name != old_name) else new_name
                db_type = await self.get_type(field.field_type.value_type)
                create_tmp += f", {db_name}  {db_type}"

            create_tmp += ');'

            copy_data = 'INSERT INTO _ax_tmp SELECT guid, axState'
            for field in ax_form.db_fields:
                copy_data += f", {field.db_name}"
            copy_data += f" FROM {ax_form.db_name};"

            drop_old = f"DROP TABLE {ax_form.db_name};"

            rename_tmp = f'ALTER TABLE _ax_tmp RENAME TO {ax_form.db_name};'

            db_session.execute(drop_tmp)
            db_session.execute(create_tmp)
            db_session.execute(copy_data)
            db_session.execute(drop_old)
            db_session.execute(rename_tmp)

        except Exception:
            logger.exception('Error executing SQL - rename_column')
            raise

    async def drop_column(self, db_session, table, column) -> None:
        """Sqlite does not allows to drop columns.
        We have to reacreate table and copy data"""
        try:
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == table
            ).first()

            drop_tmp = f"DROP TABLE IF EXISTS _ax_tmp"
            create_tmp = (
                f"CREATE TABLE _ax_tmp ("
                f"guid VARCHAR, "
                f"axState VARCHAR"
            )
            for field in ax_form.db_fields:
                db_type = await self.get_type(field.field_type.value_type)
                if field.db_name != column:
                    create_tmp += f", {field.db_name}  {db_type}"
            create_tmp += ');'

            copy_data = 'INSERT INTO _ax_tmp SELECT guid, axState'
            for field in ax_form.db_fields:
                if field.db_name != column:
                    copy_data += f", {field.db_name}"
            copy_data += f" FROM {ax_form.db_name};"
            drop_old = f"DROP TABLE {ax_form.db_name};"
            rename_tmp = f'ALTER TABLE _ax_tmp RENAME TO {ax_form.db_name};'

            db_session.execute(drop_tmp)
            db_session.execute(create_tmp)
            db_session.execute(copy_data)
            db_session.execute(drop_old)
            db_session.execute(rename_tmp)

            return True
        except Exception:
            logger.exception('Error executing SQL - drop_column')
            raise


class MysqlDialect(object):
    """SQL query for Mysql database"""

    name = 'mysql'
    version = None

    async def get_type(self, type_name):
        """Get dialect specific type"""
        mysql_types = {
            'TEXT': 'TEXT',
            'VARCHAR(255)': 'VARCHAR(255)',
            'INT': 'INT',
            'DECIMAL(65,2)': 'DECIMAL(65,2)',
            'BOOL': 'BOOL',
            'GUID': 'CHAR(32)',
            'JSON': 'TEXT',
            'TIMESTAMP': 'TIMESTAMP',
            'BLOB': 'BLOB'
        }
        return mysql_types[type_name]


    async def get_value(self, type_name, value):
        """ SELECT cat return different types for diferent dialects
        This method makes returning value same """
        del type_name
        ret_value = value
        return ret_value

    async def create_data_table(self, db_session, db_name) -> None:
        """Create table with system columns"""
        try:
            sql = f"""CREATE TABLE {db_name} (
                guid VARCHAR(64) PRIMARY KEY,
                axState VARCHAR(255) NOT NULL
            );"""
            db_session.execute(sql)
        except Exception:
            logger.exception('Error executing SQL - create_data_table')
            raise


def init_dialects(_dialect_name: str) -> None:
    """Initiate sql queries for right dialect"""
    try:
        this.dialect_name = _dialect_name
        if _dialect_name == 'sqlite':
            this.dialect = SqliteDialect()
            this.dialect.version = sqlite3.sqlite_version
        elif _dialect_name == 'postgre':
            this.dialect = PorstgreDialect()
        elif _dialect_name == 'mysql':
            this.dialect = MysqlDialect()
        else:
            raise Exception('Unknown dialect = ' + _dialect_name)

        logger.debug("Dialect initialised -> {dialect} -> {version}",
                     dialect=this.dialect.name, version=this.dialect.version)

    except Exception:
        logger.exception('Error initiating dialects')
        raise



""" This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Contains GQL schema for form actions - DoAction, subscribtion, avalible actions
"""
import os
import sys
import uuid
import copy
import asyncio
# import sys
import traceback
from concurrent.futures import ThreadPoolExecutor
from functools import partial
import graphene
import aiopubsub
from dotmap import DotMap
from loguru import logger
import ujson as json

from backend.model import AxForm, AxAction
import backend.model as ax_model
import backend.dialects as ax_dialects
import backend.pubsub as ax_pubsub
import backend.misc as ax_misc
import backend.scheduler as ax_scheduler

import backend.cache as ax_cache
from backend.schemas.types import Action, Form, ConsoleMessage, \
    ActionNotifyMessage

import backend.fields.Ax1tom as AxFieldAx1tom   # pylint: disable=unused-import
import backend.fields.Ax1tomTable as AxFieldAx1tomTable  # pylint: disable=unused-import
import backend.fields.AxChangelog as AxFieldAxChangelog  # pylint: disable=unused-import
import backend.fields.AxFiles as AxFieldAxFiles  # pylint: disable=unused-import
import backend.fields.AxImageCropDb as AxFieldAxImageCropDb  # pylint: disable=unused-import
import backend.fields.AxNum as AxFieldAxNum  # pylint: disable=unused-import


this = sys.modules[__name__]
action_loop = None


class InterpreterError(Exception):
    """ Class for AxAction python code errors """


class Executor:
    """In most cases, you can just use the 'execute' instance as a
    function, i.e. y = await execute(f, a, b, k=c) => run f(a, b, k=c) in
    the executor, assign result to y. The defaults can be changed, though,
    with your own instantiation of Executor, i.e. execute =
    Executor(nthreads=4)"""

    def __init__(self, loop, nthreads=1):
        self._ex = ThreadPoolExecutor(nthreads)
        self._loop = loop

    def __call__(self, func, *args, **kw):
        result = self._loop.run_in_executor(
            self._ex, partial(func, *args, **kw))
        return result



async def console_log(msg, modal_guid):
    """ Method used in AxAction terminal. It sends message to vue via ws """
    ax_pubsub.publisher.publish(
        aiopubsub.Key('console_log'),
        {
            'text': msg,
            'modal_guid': modal_guid
        })
    await asyncio.sleep(0.01)


class ConsoleSender:
    """ Used to send ax_console web socket messages to AxForm"""
    def __init__(self, modal_guid):
        self.modal_guid = modal_guid

    def __call__(self, msg):
        console_func = partial(console_log, msg=msg, modal_guid=self.modal_guid)
        this.action_loop.call_soon_threadsafe(console_func)
        asyncio.ensure_future(console_log(msg=msg, modal_guid=self.modal_guid),
                              loop=this.action_loop)


class ActionExecuter:
    """ Runs execute_action with pre-set db_session """
    def __init__(self, db_session):
        self.db_session = db_session

    def __call__(
            self,
            row_guid=None,
            form_guid=None,
            form_db_name=None,
            action_guid=None,
            action_db_name=None,
            values=None,
            modal_guid=None,
            arguments=None):
        return this.execute_action(
            db_session=self.db_session,
            row_guid=row_guid,
            form_guid=form_guid,
            form_db_name=form_db_name,
            action_guid=action_guid,
            action_db_name=action_db_name,
            values=values,
            modal_guid=modal_guid,
            arguments=arguments
        )


async def get_actions(form, current_state=None):
    """ Get actions for current state

    Args:
        form (AxForm): Current form
        current_state (str, optional): Name of current state.
            If None, than it is create action. Defaults to None.

    Returns:
        List(AxAction): List of avalible actions for current form and state
    """
    # TODO: Permissions filter actions, filter fields based on state and user
    ret_actions = []
    state_guid = None
    all_state_guid = None

    for state in form.states:
        if current_state:
            if state.name == current_state:
                state_guid = state.guid
        else:
            if state.is_start:
                state_guid = state.guid
        if state.is_all:
            all_state_guid = state.guid

    for action in form.actions:
        if action.from_state_guid == state_guid:
            ret_actions.append(action)
        if current_state and action.from_state_guid == all_state_guid:
            ret_actions.append(action)
    return ret_actions


async def aexec(code, localz, **kwargs):
    """ This function wraps python code from AxAction with async function
    and runs it """
    # Restore globals later
    args = ", ".join(list(kwargs.keys()))
    code_from_action = code.replace("\n", "\n    ")
    async_code = (f"async def func({args}):"
                  f"\n    {code_from_action}"
                  f"\n    return ax")
    exec(async_code, {}, localz)    # pylint: disable=exec-used
    # Don't expect it to return from the coro.
    result = await localz["func"](**kwargs)
    return result



async def do_exec(db_session, action, form, arguments=None, modal_guid=None):
    """ Executes python commands form AxAction.code

    Args:
        action (AxAction): Current action that is performed
        form (AxForm): Tobe form. Its same as action.form, but it contains
            values from vue form
        arguments(**kargs): Custom dict that can be used in code
        modal_guid (str): AxForm modal guid for wss console subscribtion

    Returns:
        Dict: Its custom Dict containing:
            info (str): Message that must be displayed to user
            error (str): Error that must be displayed to user
            item (DotMap): Javascript style dictionary of field and values.
                example- item.guid -> guid of current row
            exception (Dict): Info on exception that accured
            abort (Bool): If set to True - the action will be aborted and row
                state will not be changed
    """
    import backend.schema as ax_schema

    # async def console_print(msg):
    #     ax_pubsub.publisher.publish(
    #         aiopubsub.Key('console_log'),
    #         {
    #             'text': msg,
    #             'modal_guid': modal_guid
    #         })
    #     await asyncio.sleep(1)

    localz = dict()
    ax = DotMap()  # javascript style dicts item['guid'] == item.guid
    ax.row.guid = form.row_guid
    ax.arguments = arguments
    ax.form = form
    ax.action = action
    ax.schema = ax_schema.schema
    ax.sql = ax_dialects.dialect.custom_query
    ax.print = ConsoleSender(modal_guid=modal_guid)
    ax.paths.uploads = ax_misc.uploads_root_dir
    ax.paths.tmp = ax_misc.tmp_root_dir
    ax.add_action_job = ax_scheduler.add_action_job
    ax.do_action = ActionExecuter(db_session=db_session)
    ax.modal_guid = modal_guid
    for field in form.db_fields:
        ax.row[field.db_name] = field.value
    localz['ax'] = ax
    line_number = None

    try:
        # exec(str(action.code), globals(), localz)   # pylint: disable=exec-used
        # await eval('exec(str(action.code), globals(), localz)')
        ret_ax = await aexec(code=str(action.code), localz=localz, ax=ax)
        ret_ax = localz['ax']

        ret_data = {
            "info": ret_ax.message if 'message' in ret_ax else None,
            "error": ret_ax.error if 'error' in ret_ax else None,
            "item": ret_ax.row,
            "exception": None,
            "abort": ret_ax.abort if 'abort' in ret_ax else None
        }
        return ret_data
    except SyntaxError as err:
        logger.error(err)
        error_class = err.__class__.__name__
        detail = err.args[0]
        line_number = err.lineno
        ret_data = {
            "info": None,
            "item": None,
            "error": None,
            "exception": {
                "error_class": error_class,
                "line_number": line_number,
                "action_name": action.name,
                "detail": detail
            },
            "abort": True
        }
        return ret_data
    except Exception as err:    # pylint: disable=broad-except
        logger.error(err)
        error_class = err.__class__.__name__
        detail = err.args[0]
        cl, exc, tb = sys.exc_info()
        del cl, exc
        # line_number = traceback.extract_tb(tb)[-1][1]
        line_number = 0
        traces = traceback.extract_tb(tb)
        for idx, trace in enumerate(traces):
            if trace.filename == '<string>':
                line_number = traceback.extract_tb(tb)[idx].lineno
        del tb

        ret_data = {
            "info": None,
            "item": None,
            "error": None,
            "exception": {
                "error_class": error_class,
                "line_number": line_number,
                "action_name": action.name,
                "detail": detail
            },
            "abort": True
        }
        return ret_data


async def get_before_form(db_session, row_guid, form, ax_action):
    """Constructs before_form AxForm object and populate fields with
    values of specific row. AxForm.fields[0].value

    Args:
        row_guid (str): guid of specific row
        form (AxForm): empty AxForm
        ax_action (AxAction): AxAction wich is currently happening

    Returns:
       before_object (AxForm), tobe_object(AxForm): before_object is AxForm
       with field values of specific row. tobe_row is dummy for future
       population"""
    tobe_form = form

    if row_guid:
        fields_names = [field for field in tobe_form.db_fields]
        before_result = await ax_dialects.dialect.select_one(
            db_session=db_session,
            form=tobe_form,
            fields_list=fields_names,
            row_guid=row_guid)

        if not before_result:
            raise Exception(
                'Error in DoAction. Cant find row for action')

        tobe_form.current_state_name = before_result[0]['axState']
        tobe_form.row_guid = before_result[0]['guid']

        if tobe_form.current_state_name != ax_action.from_state.name \
                and ax_action.from_state.is_all is False:
            raise Exception('Error in DoAction. Performed \
            action does not fit axState')

        # populate each AxField of before_form with old data
        for field in tobe_form.db_fields:
            if field.db_name in before_result[0].keys():
                field.value = before_result[0][field.db_name]
                if (field.value and isinstance(field.value, str) and
                        field.field_type.value_type == 'JSON'):
                    field.value = json.loads(field.value)

    before_form = copy.deepcopy(tobe_form)
    return before_form, tobe_form


async def run_field_backend(db_session, field, action, before_form, tobe_form,
                            current_user, when='before', query_type='update'):
    """Some field types need backend execution on insert/update/delete

    Args:
        field (AxField): contains current .value of updated field
        action (AxAction): Just for info
        before_form (AxForm): Just for info
        tobe_form (AxForm): Just for info
        current_user (AxUser): Just for info
        when (str, optional): Can be 'before' or 'after'.
        query_type (str, optional): Can be 'insert', 'update', 'delete'.
        Defaults to 'update'.

    Returns:
        object: Returns new field value
    """
    tag = field.field_type.tag
    field_py_file_path = f"backend/fields/{tag}.py"
    function_name = f"{when}_{query_type}"

    if os.path.exists(ax_misc.path(field_py_file_path)):
        field_py = globals()[f'AxField{tag}']
        if field_py and hasattr(field_py, function_name):
            method_to_call = getattr(field_py, function_name)
            new_value = await method_to_call(
                db_session=db_session,
                field=field,
                before_form=before_form,
                tobe_form=tobe_form,
                action=action,
                current_user=current_user)
            return new_value
    return field.value


async def execute_action(
        db_session,
        row_guid=None,
        form_guid=None,
        form_db_name=None,
        action_guid=None,
        action_db_name=None,
        values=None,
        modal_guid=None,
        arguments=None):
    """ Performs Action on row
        # 0. Get AxForm
        # 1. Get AxAction and query_type
        # 2. Get before_form - fill it with values from database row
        # 3. Assemble tobe_object - fill it with values from AxForm.vue
        # 4. Run before backend code for each field.
            FieldType can have backend code wich is performed before and after
            action. Sea backend.fields for more info.
        # 5. Run python action, rewrite tobe_object
        # 6. Make update or insert or delete query. Db commit is here.
        # 7. Run after backend code for each field.
        # 8. Fire all web-socket subscribtions. Notify of action performed
            AxForm.vue will display message, that current row have been modified
            AxGrid.vue will reload data of grid

    Args:
        row_guid (str): Guid of row that is opened
        form_guid (str): Guid of AxForm that is opened
        form_db_name (str): db_name of AxForm that is opened.
            (Form guid or db_name must be specified)
        action_guid (str): Guid of AxAction that is performed
        action_db_name (str): db_name of AxAction
            (Action guid or db_name must be specified)
        values (Dict): Dict from AxForm.vue with values of form
        modal_guid (str): Guid generated by AxForm.vue. It used in web-socket
            subscribtion. If current form is performing this action - it does
            not need to notify user of performed action

    Returns:
        form (AxForm): Form that is changed by action. The field values could
            have been changed, if row is created - the row_guid is changed,
            state of current row changed.
        new_guid (str): Guid of current row. Null if row deleted
        messages (Dict): Custom dict containing messages, errors, exceptions
            from executing python code of action. Sea do_exec func for info.
        modal_guid (str): Same as Args. It used in web-socket
            subscribtion. If current form is performing this action - it does
            not need to notify user of performed action
        ok (bool): Used in gql query
    """
    try:
        err = 'action_schema -> execute_action'
        with ax_model.try_catch(db_session, err) as db_session:
            new_guid = uuid.uuid4()
            if not values:
                values = {}
            current_user = None  # TODO: implement users

            # Check if row locked, if not -> lock it
            if row_guid:
                lock_key = f'lock_{row_guid}'
                is_locked = await ax_cache.cache.get(lock_key)
                if not is_locked:
                    await ax_cache.cache.set(lock_key, True)
                else:
                    messages = {
                        "exception": None,
                        "error": "$i18n('form.row-locked')",
                        "info": None
                    }
                    messages_json = json.dumps(messages)
                    return {
                        "form": None,
                        "new_guid": None,
                        "messages": messages_json,
                        "modal_guid": modal_guid,
                        "ok": False
                    }

            # 0. Get AxForm
            ax_form = None
            if form_guid:
                ax_form = db_session.query(AxForm).filter(
                    AxForm.guid == uuid.UUID(form_guid)
                ).first()
            elif form_db_name:
                ax_form = db_session.query(AxForm).filter(
                    AxForm.db_name == form_db_name
                ).first()
            else:
                raise ValueError(
                    'doAction GQL - form guid or db_name required')

            db_session.expire(ax_form)

            # 1. Get AxAction and query_type
            ax_action = None
            if not action_guid and not action_db_name:
                raise ValueError(
                    'doAction GQL - action guid or db_name required')
            for action in ax_form.actions:
                if not ax_action:
                    if action_guid and action.guid == uuid.UUID(action_guid):
                        ax_action = action
                    if action_db_name and action.db_name == action_db_name:
                        ax_action = action

            query_type = 'update'
            if ax_action.from_state.is_start:
                query_type = 'insert'
            elif ax_action.to_state.is_deleted:
                query_type = 'delete'

            # 2. Get before_form - fill it with values from database row
            before_form, tobe_form = await get_before_form(
                db_session=db_session,
                row_guid=row_guid,
                form=ax_form,
                ax_action=ax_action)

            # 3. Assemble tobe_object - fill it with values from AxForm.vue
            if not tobe_form.row_guid:
                tobe_form.row_guid = new_guid
            for field in tobe_form.db_fields:
                if field.db_name in values.keys():
                    field.value = values[field.db_name]
                    field.needs_sql_update = True
                if field.field_type.is_updated_always:
                    field.needs_sql_update = True

            # 4. Run before backend code for each field.
            for field in tobe_form.no_tab_fields:
                if field.field_type.is_backend_available:
                    field.value = await run_field_backend(
                        db_session=db_session,
                        when='before',
                        query_type=query_type,
                        field=field,
                        action=ax_action,
                        before_form=before_form,
                        tobe_form=tobe_form,
                        current_user=current_user)

            # 5. Run python action, rewrite tobe_object
            messages = None
            messages_json = None
            if ax_action.code:
                code_result = await do_exec(
                    db_session=db_session,
                    action=ax_action,
                    form=tobe_form,
                    arguments=arguments,
                    modal_guid=modal_guid)
                messages = {
                    "exception": code_result["exception"],
                    "error": code_result["error"],
                    "info": code_result["info"]
                }
                messages_json = json.dumps(messages)
                if code_result['abort'] or code_result["exception"]:
                    await ax_cache.cache.delete(f'lock_{row_guid}')
                    return {
                        "form": before_form,
                        "new_guid": None,
                        "messages": messages_json,
                        "modal_guid": modal_guid,
                        "ok": False
                    }
                # update fields with values recieved from actions python
                new_item = code_result['item']
                for field in tobe_form.db_fields:
                    if field.db_name in new_item.keys():
                        field.value = new_item[field.db_name]
                        field.needs_sql_update = True

            # 6. Make update or insert or delete query #COMMIT HERE
            return_guid = tobe_form.row_guid
            after_form = copy.deepcopy(tobe_form)
            if query_type == 'insert':
                await ax_dialects.dialect.insert(
                    db_session=db_session,
                    form=tobe_form,
                    to_state_name=ax_action.to_state.name,
                    new_guid=new_guid
                )
                tobe_form.row_guid = new_guid
                return_guid = new_guid
            elif query_type == 'delete':
                await ax_dialects.dialect.delete(
                    db_session=db_session,
                    form=tobe_form,
                    row_guid=tobe_form.row_guid
                )
                return_guid = None
            else:
                await ax_dialects.dialect.update(
                    db_session=db_session,
                    form=tobe_form,
                    to_state_name=ax_action.to_state.name,
                    row_guid=tobe_form.row_guid
                )

            # 7. Run after backend code for each field.
            for field in after_form.no_tab_fields:
                if field.field_type.is_backend_available:
                    field.value = await run_field_backend(
                        db_session=db_session,
                        when='after',
                        query_type=query_type,
                        field=field,
                        action=ax_action,
                        before_form=before_form,
                        tobe_form=after_form,
                        current_user=current_user)

            # 8. Fire all web-socket subscribtions. Notify of action performed
            # subscription_form = AxForm()
            # subscription_form.guid = tobe_form.guid
            # subscription_form.icon = tobe_form.icon
            # subscription_form.db_name = tobe_form.db_name
            # subscription_form.row_guid = tobe_form.row_guid
            # subscription_form.modal_guid = modal_guid

            notify_message = {
                "form_guid": tobe_form.guid,
                "form_icon": tobe_form.icon,
                "form_db_name": tobe_form.db_name,
                "row_guid": tobe_form.row_guid,
                "modal_guid": modal_guid,
                "action_guid": ax_action.guid,
                "action_db_name": ax_action.db_name,
                "action_icon": ax_action.icon
            }

            ax_pubsub.publisher.publish(
                aiopubsub.Key('do_action'), notify_message)

            await ax_cache.cache.delete(f'lock_{row_guid}')
            return {
                "form": tobe_form,
                "new_guid": return_guid,
                "messages": messages_json,
                "modal_guid": modal_guid,
                "ok": True
            }
    except Exception:
        await ax_cache.cache.delete(f'lock_{row_guid}')
        logger.exception('Error while executing action - DoAction.')
        raise


def run_async(corofn, *args):
    """  run_in_executor takes only sync function, but with
    this hack we can run corutines """
    loop = asyncio.new_event_loop()
    coro = corofn(*args)
    asyncio.set_event_loop(loop)
    return loop.run_until_complete(coro)


async def run_execute_action(
        db_session,
        row_guid=None,
        form_guid=None,
        form_db_name=None,
        action_guid=None,
        action_db_name=None,
        values=None,
        modal_guid=None,
        arguments=None
    ):
    """ This function executes execute_action corutine in different thread """
    loop = asyncio.get_event_loop()
    executor = ThreadPoolExecutor(max_workers=5)
    execute_func = partial(
        execute_action,
        db_session=db_session,
        row_guid=row_guid,
        form_guid=form_guid,
        form_db_name=form_db_name,
        action_guid=action_guid,
        action_db_name=action_db_name,
        values=values,
        modal_guid=modal_guid,
        arguments=arguments
    )

    result = await loop.run_in_executor(
        executor,
        run_async,
        execute_func)

    return result




class DoAction(graphene.Mutation):
    """
        Runs execute_action. Performs action on row.
    """
    class Arguments:  # pylint: disable=missing-docstring
        row_guid = graphene.String(required=False, default_value=None)
        form_guid = graphene.String(required=False, default_value=None)
        form_db_name = graphene.String(required=False, default_value=None)
        action_guid = graphene.String(required=False, default_value=None)
        action_db_name = graphene.String(required=False, default_value=None)
        values = graphene.String()
        arguments = graphene.String(required=False, default_value=None)
        modal_guid = graphene.String(required=False, default_value=None)

    form = graphene.Field(Form)
    new_guid = graphene.String()
    messages = graphene.String()
    modal_guid = graphene.String()
    ok = graphene.Boolean()

    async def mutate(self, info, **args):  # pylint: disable=missing-docstring
        err = 'Error in gql mutation - action_schema -> DoAction.'
        with ax_model.try_catch(info.context['session'], err) as db_session:
            values_string = args.get('values')
            values = {}
            if values_string and values_string != 'null':
                values = json.loads(values_string)
            row_guid = args.get('row_guid')
            modal_guid = args.get('modal_guid')
            form_guid = args.get('form_guid')
            form_db_name = args.get('form_db_name')
            action_guid = args.get('action_guid')
            action_db_name = args.get('action_db_name')
            arguments_string = args.get('arguments')
            arguments = {}
            if arguments_string and arguments_string != 'null':
                arguments = json.loads(arguments_string)

            result = await run_execute_action(
                db_session=db_session,
                row_guid=row_guid,
                form_guid=form_guid,
                form_db_name=form_db_name,
                action_guid=action_guid,
                action_db_name=action_db_name,
                values=values,
                arguments=arguments,
                modal_guid=modal_guid)

            return DoAction(
                form=result["form"],
                new_guid=result["new_guid"],
                messages=result["messages"],
                modal_guid=result["modal_guid"],
                ok=result["ok"])


class ActionSubscription(graphene.ObjectType):
    """GraphQL subscriptions"""
    action_notify = graphene.Field(
        ActionNotifyMessage,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        row_guid=graphene.Argument(type=graphene.String, required=False))

    async def resolve_action_notify(self, info, form_db_name, row_guid=None):
        """ Web-socket subscription on every performed action
            AxForm.vue will display message, that current row have been modified
            AxGrid.vue will reload data of grid

            Args:
                form_db_name (str): db_name of AxForm
                row_guid (str): If it is set, then the subscriber will be
                    notified only on actions performed with surtain row.
                    AxGrid is notified on every action
                    AxForm is notified only on row actions

            Returns:
                payload (Dict): Dict containing action info:
                    form_guid (str): AxForm.guid of current form
                    form_icon (str): font-awesome icon
                    form_db_name (str): AxForm.db_name of current form
                    row_guid (str): Guid of row on wich action is performed
                    modal_guid (str): Guid generated by AxForm.vue.
                        It used in web-socket subscribtion. If current form is
                        performing this action - it doesnot need to notify user
                        of performed action
                    action_guid (str): Current action guid
                    action_db_name (str): Db name of current action
                    action_icon (str): font-awesome icon of current action

        """
        del info
        try:
            subscriber = aiopubsub.Subscriber(
                ax_pubsub.hub, 'action_notify_subscriber')
            subscriber.subscribe(aiopubsub.Key('do_action'))
            while True:
                key, payload = await subscriber.consume()
                del key
                if payload['form_db_name'] == form_db_name:
                    if row_guid is None or row_guid == payload['row_guid']:
                        message = ActionNotifyMessage(
                            form_guid=payload['form_guid'],
                            form_icon=payload['form_icon'],
                            form_db_name=payload['form_db_name'],
                            row_guid=payload['row_guid'],
                            modal_guid=payload['modal_guid'],
                            action_guid=payload['action_guid'],
                            action_db_name=payload['action_db_name'],
                            action_icon=payload['action_icon'])
                        yield message
        except asyncio.CancelledError:
            await subscriber.remove_all_listeners()


    console_notify = graphene.Field(
        ConsoleMessage,
        modal_guid=graphene.Argument(type=graphene.String, required=True))

    async def resolve_console_notify(self, info, modal_guid):
        """ Web-socket subscription on ax terminal print
        """
        del info
        try:
            subscriber = aiopubsub.Subscriber(
                ax_pubsub.hub, 'action_notify_subscriber')
            subscriber.subscribe(aiopubsub.Key('console_log'))
            this.action_loop = asyncio.get_event_loop()
            while True:
                key, payload = await subscriber.consume()
                await asyncio.sleep(0.1)
                del key
                if payload['modal_guid'] == modal_guid:
                    message = ConsoleMessage(
                        text=payload['text'],
                        modal_guid=payload['modal_guid'])
                    yield message
        except asyncio.CancelledError:
            await subscriber.remove_all_listeners()
        except Exception: # pylint: disable=broad-except
            logger.exception('Error in gql sub resolve_console_notify.')



class ActionQuery(graphene.ObjectType):
    """Query all data of AxAction and related classes"""
    action_data = graphene.Field(
        Action,
        guid=graphene.Argument(type=graphene.String, required=True),
        update_time=graphene.Argument(type=graphene.String, required=False))
    actions_avalible = graphene.List(
        Action,
        form_db_name=graphene.Argument(type=graphene.String, required=True),
        current_state=graphene.Argument(type=graphene.String, required=False),
        update_time=graphene.Argument(type=graphene.String, required=False))

    async def resolve_action_data(self, info, guid=None, update_time=None):
        """ Gets AxAction by guid. Used in TheActionModal.vue. """
        del update_time
        err = 'action_schema -> resolve_action_data'
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True):
            query = Action.get_query(info=info)
            ax_action = query.filter(AxAction.guid == uuid.UUID(guid)).first()
            return ax_action

    async def resolve_actions_avalible(self, info, form_db_name,
                                       current_state=None, update_time=None):
        """ Gets AxActions for current state """
        # TODO: Add permissions based on roles and current user
        del update_time
        err = (f"action_schema.resolve_actions_avalible."
               f"Error in gql query - resolve_actions_avalible.")
        with ax_model.try_catch(
                info.context['session'], err, no_commit=True) as db_session:
            ax_form = db_session.query(AxForm).filter(
                AxForm.db_name == form_db_name
            ).first()
            ax_actions = await get_actions(
                form=ax_form, current_state=current_state)

            return ax_actions



class ActionMutations(graphene.ObjectType):
    """Combine all mutations"""
    do_action = DoAction.Field()



"""Module with aiocache instance. Works inMemory or with redis
"""

import sys
from aiocache import caches
from loguru import logger

this = sys.modules[__name__]
cache = None


def init_cache(
        mode: str,
        redis_endpoint: str,
        redis_port: int,
        redis_timeout: int) -> bool:
    """Initiate aiocache cache.

    Args:
        mode (str, optional): Defaults to 'default'. Set 'redis' for Redis.
        redis_endpoint (str, optional): Defaults to '127.0.0.1'. IP of Redis.
        redis_port (int, optional): Defaults to 6379. Port of Redies server.
        redis_timeout (int, optional): Defaults to 1. Redis timeout.

    Returns:
        bool: True if succesuful
    """

    if mode.lower == 'redis':
        redis_config = {
            'redis': {
                'cache': "aiocache.RedisCache",
                'endpoint': redis_endpoint,
                'port': redis_port,
                'timeout': redis_timeout,
                'serializer': {
                    'class': "aiocache.serializers.PickleSerializer"
                },
                'plugins': [
                    {'class': "aiocache.plugins.HitMissRatioPlugin"},
                    {'class': "aiocache.plugins.TimingPlugin"}
                ]
            }
        }
        # TODO check if Redis is up and running with given settings

        try:
            caches.set_config(redis_config)
            this.cache = caches.get('redis')
        except Exception:
            logger.exception('Error initiating aiocache with Redis.')
            raise
    else:
        ram_config = {
            'default': {
                'cache': "aiocache.SimpleMemoryCache",
                'serializer': {
                    'class': "aiocache.serializers.PickleSerializer"
                }
            }
        }

        try:
            caches.set_config(ram_config)
            this.cache = caches.get('default')
        except Exception:
            logger.exception(
                'Error initiating aiocache with SimpleMemoryCache. ')
            raise

    return True



"""This is part of GraphQL schema (Mutaions, Queryes, Subscriptions).
Defines Fields Scheme and all mutations.
Will be used for mutations used in field_types"""

import graphene
from graphene_sqlalchemy.converter import convert_sqlalchemy_type
from backend.misc import convert_column_to_string
from backend.model import GUID

convert_sqlalchemy_type.register(GUID)(convert_column_to_string)


class MutationExample(graphene.Mutation):
    """Mutation example"""
    class Arguments:  # pylint: disable=missing-docstring
        input_text = graphene.String()

    output_text = graphene.String()

    async def mutate(self, info, input_text):  # pylint: disable=missing-docstring
        del info
        return MutationExample(output_text=input_text)


class FieldsQuery(graphene.ObjectType):
    """AxUser queryes"""


class FieldsMutations(graphene.ObjectType):
    """Contains all AxUser mutations"""
    mutation_example = MutationExample.Field()



import sys
import os
import time
import atexit
import signal


# class AxDaemon(daemon):
#     def run(self):
#         print(ax_logo)
#         logger.info(
#             f'Running Ax on {self.host}:{self,port}, {self.workers} workers running')
#         app.run(
#             host=self.host,
#             port=self.port,
#             debug=self.debug,
#             access_log=self.access_log,
#             workers=self.workers)

# daemon = AxDaemon(
#     pidfile='ax-daemon.pid',
#     host=host,
#     port=port,
#     debug=debug,
#     access_log=access_log,
#     workers=workers
# )

# if arguments['start']:
#     daemon.start()

# if arguments['stop']:
#     daemon.stop()

# if arguments['restart']:
#     daemon.restart()


class daemon:
    """A generic daemon class.

    Usage: subclass the daemon class and override the run() method."""

    def __init__(self, pidfile, host, port, debug, access_log, workers):
        self.pidfile = pidfile
        self.host = host
        self.port = port
        self.debug = debug
        self.access_log = access_log
        self.workers = workers

    def daemonize(self):
        """Deamonize class. UNIX double fork mechanism."""

        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #1 failed: {0}\n'.format(err))
            sys.exit(1)

        # decouple from parent environment
        os.chdir('/')
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:

                # exit from second parent
                sys.exit(0)
        except OSError as err:
            sys.stderr.write('fork #2 failed: {0}\n'.format(err))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = open(os.devnull, 'r')
        so = open(os.devnull, 'a+')
        se = open(os.devnull, 'a+')

        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)

        pid = str(os.getpid())
        with open(self.pidfile, 'w+') as f:
            f.write(pid + '\n')

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """Start the daemon."""

        # Check for a pidfile to see if the daemon already runs
        try:
            with open(self.pidfile, 'r') as pf:

                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if pid:
            message = "pidfile {0} already exist. " + \
                "Daemon already running?\n"
            sys.stderr.write(message.format(self.pidfile))
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """Stop the daemon."""

        # Get the pid from the pidfile
        try:
            with open(self.pidfile, 'r') as pf:
                pid = int(pf.read().strip())
        except IOError:
            pid = None

        if not pid:
            message = "pidfile {0} does not exist. " + \
                "Daemon not running?\n"
            sys.stderr.write(message.format(self.pidfile))
            return  # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError as err:
            e = str(err.args)
            if e.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print(str(err.args))
                sys.exit(1)

    def restart(self):
        """Restart the daemon."""
        self.stop()
        self.start()

    def run(self):
        """You should override this method when you subclass Daemon.

        It will be called after the process has been daemonized by 
        start() or restart()."""
